package mygroup.myartifact;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.UUID;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

import mygroup.myartifact.dispatchers.Handler;
import mygroup.myartifact.modules.ModuleManager;
import mygroup.myartifact.modules.Node;
import mygroup.myartifact.modules.NodeInfo;
import mygroup.myartifact.modules.messages.BroadcastDispatchers;
import mygroup.myartifact.modules.messages.BroadcastNickname;
import mygroup.myartifact.modules.messages.BroadcastNodeInfo;
import mygroup.myartifact.modules.messages.NodeMessage;
import mygroup.myartifact.modules.messages.NodeResponseMessage;
import mygroup.myartifact.modules.messages.TellNodeInfo;
import mygroup.myartifact.modules.messages.UnicastMessage;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Distinguish between node-level and module-level. Nodes interact and change messages with node-level messages on top
 * of transport mechanism offered by modules. Distinguish between node-specific information and module-specific
 * information. Nodes exchange information like ID, nick, and available handler. Modules are freed from managing this
 * information, but only deal with information needed to find another nodes.
 * 
 * @author dell
 * 
 */
public class MachineNode {
	public static final String LIBRARY_NAME = "HEKATE"; // PROJECT_NAME, Names for die Awareness-Library:

	private Logger _logger;

	private Properties _properties;
	private ModuleManager _moduleManager;
	private String _id;
	private String _nick;
	private List<Handler> _handlers;
	
	private boolean _running;
	
	private List<NodeRequest> _nodeRequests; // Manage request-response mechanism in MachineNode for node messages
	
	public MachineNode(String propertiesFile) {
		_logger = LoggerFactory.getLogger(MachineNode.class);
		_properties = new Properties();
		_running = false;
		
		if(propertiesFile==null) { 
			propertiesFile = "/mygroup/myartifact/configurations/webstart.properties"; // Default configuration
		}
		
		// First try to from local file, if this fails try to load from class-path, if this fails load default configuration file
		try {
			// Java Web Start doesn't like this line (java.security.AccessControlException), seems like a SecurityManager with very restrictive local permissions is loaded 	
			_properties.load(new FileReader(new File(propertiesFile)));
			_logger.debug("Loaded configuration from local file.");
		} catch (Exception e1) {
			_logger.debug("Could not read configuration from local file, trying to load from class path...");
			try {
				_properties.load(MachineNode.class.getResourceAsStream(propertiesFile));
				_logger.debug("Loaded configuration "+propertiesFile+" from class path.");
			} catch (IOException e) {
				_logger.debug("Could not find configuration file "+propertiesFile+", using default/random values:");
				try {
					_properties.load(MachineNode.class.getResourceAsStream("/mygroup/myartifact/configurations/webstart.properties"));
				} catch (IOException e3) {
					_logger.error("This should never happen, programm in undefined state.");
					e3.printStackTrace();
				}
			}
		}
		
		_nodeRequests = new ArrayList<NodeRequest>();
		
		// How to ensure unique *and* persistent IDs between distributed nodes? To ensure 100% unique IDs, a central instance is needed generating unique IDs.
		// Let central instance choose unique ID? It this a good idea in a distributed node network? Generally no!
		// See well-known theoretical results from distributed algorithms, there needs to be some kind of pre-defined value to distinguish between nodes.
		// For the moment UUIDs should be sufficient to use.
		if(_properties.getProperty("id")!=null) {
			_id = _properties.getProperty("id");
		} else {
			_id = UUID.randomUUID().toString(); // UUID ensures unique ID (with high probability), which can be also stored and loaded persistently between different node-runtimes
		}
		
		if(_properties.getProperty("nick")!=null) {
			_nick = _properties.getProperty("nick");
		} else {
			try {
				_nick = InetAddress.getLocalHost().getHostName();
			} catch (UnknownHostException e) {
				_nick = null;
			}
		}
		
		_handlers = new ArrayList<Handler>();
		if (_properties.getProperty("dispatcher") != null) {
			for (String handlerString : _properties.getProperty("dispatcher").split(",")) {
				Class<?> cls;
				try {
					cls = (Class<?>) Class.forName(handlerString);
					
					Class superClass = cls.getSuperclass();
					while(superClass!=null) {
						if( superClass.equals(mygroup.myartifact.dispatchers.Handler.class) ) {
							try {
								//Constructor cons = cls.getConstructor(); // no-arg constructor of DispatcherFactory
								Constructor cons = cls.getConstructor(MachineNode.class, Properties.class);
								try {
									//DispatcherFactory dispatcherFactory = (DispatcherFactory) cons.newInstance();
									//Handler dispatcherObject = dispatcherFactory.newInstance(this, _properties);
									//this._handlers.add(dispatcherObject);
									Handler handler = (Handler) cons.newInstance(this, _properties);
									this._handlers.add(handler);
									_logger.debug("Handler {} loaded.", handler.getIdentifier());
									// Start dispatcher later, with GUI or another mechanism
								} catch (Exception e) { // Single Exception-block instead of four times writing the same log for IllegalArgumentException, InstantiationException, IllegalAccessException and InvocationTargetException
									_logger.debug("Could not instanciate {}.", handlerString);
									e.printStackTrace();
								}
							} catch (SecurityException e) {
								_logger.debug("Could not get default (no-arg) constructor for class {} due to security manager constraints.", handlerString);
								e.printStackTrace();
							} catch (NoSuchMethodException e) {
								_logger.debug("Could not get default (no-arg) constructor for class {}, but needed to setup dispatcher. Supply class with default constructor.", handlerString);
								e.printStackTrace();
							}
							break;
						}
						superClass = superClass.getSuperclass();
					}
					
					
					
//					if (Arrays.asList(cls.getInterfaces()).contains(mygroup.myartifact.dispatchers.DispatcherFactory.class)) { // Class needs to implement Dispatcher-interface
//						try {
//							Constructor cons = cls.getConstructor(); // no-arg constructor of DispatcherFactory
//							try {
//								DispatcherFactory dispatcherFactory = (DispatcherFactory) cons.newInstance();
//								Handler dispatcherObject = dispatcherFactory.newInstance(this, _properties); // Does -load gets stolen here?
//								this._handlers.add(dispatcherObject);
//								_logger.debug("Dispatcher {} loaded.", dispatcherObject.getIdentifier());
//								// Start dispatcher later, with GUI or another mechanism
//							} catch (Exception e) { // Single Exception-block instead of four times writing the same log for IllegalArgumentException, InstantiationException, IllegalAccessException and InvocationTargetException
//								_logger.debug("Could not instanciate {}.", handlerString);
//								e.printStackTrace();
//							}
//						} catch (SecurityException e) {
//							_logger.debug("Could not get default (no-arg) constructor for class {} due to security manager constraints.", handlerString);
//							e.printStackTrace();
//						} catch (NoSuchMethodException e) {
//							_logger.debug("Could not get default (no-arg) constructor for class {}, but needed to setup dispatcher. Supply class with default constructor.", handlerString);
//							e.printStackTrace();
//						}
//					}
					
					
					
				} catch (ClassNotFoundException e) {
					_logger.debug("Could not found specified dispatcher class {}, can't load this dispatcher", handlerString);
					e.printStackTrace();
				}
			}
		}
		
		this._moduleManager = new ModuleManager(_properties, this); // Start modules
	}
	
	public Properties getProperties() {
		return _properties;
	}
	
	/* Helper methods for other classes */
	public String getId() {
		return this._id;
	}

	public String getNick() {
		return this._nick;
	}

	public String getIdentifier() {
		return (_nick == null) ? _id : _nick;
	}

	public void setNick(String nick) {
		if(!_nick.equals(nick)) {
			this._nick = nick;
			
			// Only try to broadcast new nick to other nodes if MachineNode already started
			if(_running) {
				// Broadcast new nick to all other nodes if caller things this appropriately in the current context
				BroadcastNickname broadcastNick = new BroadcastNickname(this._id, this._nick);
				this._moduleManager.sendNodeMessage(broadcastNick); // Makes no sense to throw NodeNotKnownException or NodeNotAvailableException if trying to send broadcast message; want to make them Runtime Exceptions?	
			}
			
		}
	}
	// Overloaded version to set (new) nickname of node, but do *not* send (new) nick to other nodes
//	public void setNick(String nick) {
//		// It is useful to set nickname of MachineNode *without* broadcasting nickname, especially when MachineNode is not started or MachineNode is currently starting/booting up, where currently there is no node available
//		setNick(nick, false);
//	}

	public List<Handler> getDispatchers() {
		return this._handlers;
	}
	// TODO there is no way to dynamically add new Dispatcher (or DispatcherFactories) at runtime; this can be fixed with new method addDispatcher(Dispatcher) or with DispatcherFactory? No, setting directly Dispatcher is the way to go
	//
	// How to modify the availability-state of a Dispatcher-object? By calling a method on MachineNode, e.g. MachineNode.setDispatcherState(Dispatcher, boolean state)
	// Or by calling a method on Dispatcher itself, e.g. Dispatcher.setEnabled(boolean state); last fells to be the better approach, encapsulate state in Dispatcher itself and check for state when using dispatcher
	//
	// In this case the method MachineNode.setDispatchers(List<Dispatcher>) obsolete and wrong as well; is isn't used anyway
	// public void setDispatchers(List<Dispatcher> dispatchers) {
	// _dispatchers = dispatchers;
	// BroadcastDispatchers broadcastDispatchers = new BroadcastDispatchers(_id, getDispatchersClassNames());
	// _moduleManager.sendNodeMessage(broadcastDispatchers);
	// }

	// Need to introduce new method getAvailableDispatchersClassNames():List<String> to introduce the new Dispatcher-availabel mechanism
	public List<String> getDispatchersClassNames() {
		List<String> strings = new ArrayList<String>();
		for (Handler dispatcher : _handlers) {
			strings.add(dispatcher.getClass().getCanonicalName());
		}
		return strings;
	}

	public List<String> getEnabledDispatchersClassNames() {
		List<String> strings = new ArrayList<String>();
		for (Handler dispatcher : _handlers) {
			if (dispatcher.isEnabled()) {
				strings.add(dispatcher.getClass().getCanonicalName());
			}
		}
		return strings;
	}

	// Used by MachineNodeGUI to indicate that list of currently available Dispatchers changed and MachineNode should tell this to other nodes; TODO this is a bad approach, switch using an Observer/Listener approach
	public void enabledDispatchersChanged() {
		BroadcastDispatchers broadcastDispatchers = new BroadcastDispatchers(_id, getEnabledDispatchersClassNames());
		_moduleManager.sendNodeMessage(broadcastDispatchers);
	}

	// Used by MachineNodeGUI to print currently available modules; possible that ModuleManager becomes obsolete? Why not let Nodes objects handle *everything* direclty!?! Let's see how this developes in the future
	public ModuleManager getModuleManager() {
		return this._moduleManager;
	}

	// Start-method, should be executed in separate Thread and not in AWT-Thread etc.
	public void start() {
		_moduleManager.start();
		_running = true;
	}
	
	public void stop() {
		_running = false;
		_moduleManager.stop();
	}

	// Called from ModuleManager when new String-message received; could be NodeMessage or some other arbitrary string
	public void dispatchMessage(String msg) {
		Object object = Utility.XML.deserialize(msg.getBytes()); // Could return null, indicating that could not serialize message

		if (object == null) {
			_logger.debug("Could not handle message, content was {}", msg);
		} else {

			// MachineNode knows how to deal with NodeMessage; anything else needs to be dumped
			if (object instanceof NodeMessage) {
				NodeMessage nodeMessage = (NodeMessage) object;
				_logger.debug("Received NodeMessage {}.", nodeMessage);
				
				// Prevent MachineNode from receiving own NodeMessage, avoiding potential message-loops, inconsistent application state etc.
				if (nodeMessage.getSender().equals(_id)) {
					_logger.debug("Ignoring own NodeMessage with message-id.", nodeMessage.getMessageId());
					return;
				}
				
				// Ignore message designated for another node
				if (object instanceof UnicastMessage) {
					UnicastMessage unicastMessage = (UnicastMessage) object;
					if (!_id.equals(unicastMessage.getReceiver())) {
						_logger.debug("Ignoring unicast-NodeMessage {} with message-id {} designated for node {}.", new Object[] { unicastMessage.getClass().getCanonicalName(), unicastMessage.getMessageId(), unicastMessage.getReceiver() });
						return;
					}
				}
				
				// There are two types of message interaction: one-shot NodeMessages and Request-Response-NodeMessages
				// -One-shot messages are very simple messages without any further requirement for processing, they can be handled directly.
				// -Request-Response messages need to merged together on the side sending the request, bringing a response to the appropriate handler that sent a request message.
				// This is also the basis for more elaborate protocols, which are handled by protocol handlers.
				// A request message utilizes a message-id, and response messages reference to them to enable a message-dispatcher to bring both together
				
				// First handle NodeResponseMessages response-messages by filtering them out and try to associate them to the appropriate NodeRequest message
				if (nodeMessage instanceof NodeResponseMessage) {
					NodeResponseMessage nodeResponseMessage = (NodeResponseMessage) nodeMessage;
					_logger.debug("Received NodeResponseMessage from node {} with message-id {} trying to reference message-id {}.", new Object[] { nodeResponseMessage.getSender(), nodeResponseMessage.getMessageId(), nodeResponseMessage.getReferencedMessageId() });
					// Try to associate response to right previously send request message
					for (NodeRequest nodeRequest : _nodeRequests) {
						if (nodeRequest._messageId.equals(nodeResponseMessage.getReferencedMessageId())) {
							_logger.debug("Successfully identified response to appropriate NodeRequest object, message-id is {}.", nodeResponseMessage.getReferencedMessageId());
							nodeRequest._response = nodeResponseMessage; // Currently only UnicastMessage can be handled as NodeResponseMessage; this may change in the near future, but I don't see any need for this currently
							nodeRequest._latch.countDown(); // Count-down latch the handler that sent the request is waiting on
							// No need to remove nodeRequest object from List, because method sendNodeRequestMessage already handles this
							return;
						}
					}
					_logger.debug("Could not associate NodeRequest referenced message id {}.", nodeResponseMessage.getReferencedMessageId());
				}

				// Handle and assume one-shot messages
				
				// New node sends information about itself to introduce itself on node-level
				else if (nodeMessage instanceof BroadcastNodeInfo) {
					BroadcastNodeInfo broadcastNodeInfo = (BroadcastNodeInfo) nodeMessage;
					NodeInfo nodeInfo = broadcastNodeInfo.getNodeInfo();
					_logger.debug("Received BroadcastNodeInfo with message-id {} from remote node {}.", new Object[] { nodeMessage.getMessageId(), nodeInfo.getId() });
					// Update information on appropriate node-object in ModuleManager._nodes
					Node node = _moduleManager.getNode(nodeInfo.getId());
					if (node != null) {
						node.setNick(nodeInfo.getNick());
						node.setDispatchers(nodeInfo.getDispatchers());
					} else {
						_logger.debug("Could not set information for node {}, because didn't know this node before. Software maybe in undefined state.", nodeInfo.getId());
					}
				}

				// A remote node got notified about this new node and PUSHes it's own information into this node
				else if (nodeMessage instanceof TellNodeInfo) {
					TellNodeInfo tellNodeInfo = (TellNodeInfo) nodeMessage;
					NodeInfo nodeInfo = tellNodeInfo.getNodeInfo();
					_logger.debug("Received TellNodeInfo with message-id {} from node {}.", new Object[] { nodeMessage.getMessageId(), nodeMessage.getSender() });
					_logger.debug("NodeInfo is ID:{} and Nick:{} and Dispatchers:{}", new Object[] { nodeInfo.getId(), nodeInfo.getNick(), nodeInfo.getDispatchers() });
					Node node = _moduleManager.getNode(nodeInfo.getId());
					// Node node = _moduleManager.getNode(tellNodeInfo.getSender()); // The same as above

					if (node != null) {
						node.setNick(nodeInfo.getNick());
						node.setDispatchers(nodeInfo.getDispatchers());
					} else {
						_logger.debug("Could not set information for remote node {}, because didn't know this node before. Software maybe in undefined state.", nodeInfo.getId());
					}
				}

				// Another MachineNode changed it's nickname
				else if (nodeMessage instanceof BroadcastNickname) {
					BroadcastNickname broadcastNick = (BroadcastNickname) nodeMessage;
					_logger.debug("Received BroadcastNickname with message-id {} from node {}. Nickname is {}", new Object[] { nodeMessage.getMessageId(), nodeMessage.getSender(), broadcastNick.getNick() });
					// Could fail if null returned from mm.getNode(String id), but shouldn't happen!?!
					this._moduleManager.getNode(broadcastNick.getSender()).setNick(broadcastNick.getNick());
				}

				// Other node updated it's list of available dispatchers
				else if (nodeMessage instanceof BroadcastDispatchers) {
					BroadcastDispatchers broadcastDispatchers = (BroadcastDispatchers) nodeMessage;
					_logger.debug("Received BroadcastDispatchers with message-id {} from node {}. Dispatchers is {}", new Object[] { nodeMessage.getMessageId(), nodeMessage.getSender(), broadcastDispatchers.getDispatchers() });
					this._moduleManager.getNode(broadcastDispatchers.getSender()).setDispatchers(broadcastDispatchers.getDispatchers());
				}

				// Forward messages to local Dispatchers
				else {
					for (Handler dispatcher : _handlers) {
						_logger.debug("Give dispatcher {} a chance to handle NodeMessage {}", new Object[] { dispatcher.getIdentifier(), nodeMessage.getClass().getCanonicalName() });
						if (dispatcher.dispatchNodeMessage(nodeMessage)) {
							_logger.debug("NodeMessage from remote node {} was successfully handled by dispatcher {}.", new Object[] { nodeMessage.getSender(), dispatcher.getIdentifier() }); // Why is this executed twice!?!
							return;
						}
						_logger.debug("Dispatcher {} wasn't able to handle NodeMessage {}.", new Object[] { dispatcher.getIdentifier(), nodeMessage.getClass().getCanonicalName() });
					}
					// Found now Dispatcher able to proceed the given message, print this
					_logger.debug("No Dispatcher could handle received NodeMessage, content of plain-text of xml message was {}", msg);
				}
			}

			// Server sent anything, but a NodeMessage; dump message
			else {
				_logger.debug("Could not handle message, content was {}", msg); // Kind-a strange doubling this line; reasonable to try to merge both lines?
			}
		}
	}

	// Used by Handlers to send arbitrary NodeMessages to remote 
	public void sendNodeMessage(NodeMessage nodeMessage) {
		this._moduleManager.sendNodeMessage(nodeMessage);
	}
	
	public UnicastMessage sendNodeRequestMessage(UnicastMessage unicastMessage, long timeout, TimeUnit timeUnit) { // Only UnicastMessage possible currently
		// Remember message-id of request-UnicastMessage to later associate right response-UnicastMessage to this UnicastMessage
		NodeRequest request = new NodeRequest(unicastMessage.getMessageId());
		_nodeRequests.add(request);

		// Send message, and wait on timeout or until latch to be count-down
		sendNodeMessage(unicastMessage);
		try {
			// boolean reachedZero = request._latch.await(timeout, timeUnit); // Interesting, true or false not important to check; didn't expect this
			boolean reachedZero = request.getLatch().await(timeout, timeUnit); // Use method for more/better maintainable code
			_nodeRequests.remove(request); // Used NodeRequest object is useless now, so why not remove it to avoid memory leaks and let application in valid state
			return request.getResponse();
		} catch (InterruptedException e) {
			_logger.debug("Waiting on CountDownLatch disturbed by interrupt. This should *never* happen. Returning null and removing NodeRequest object to bring application back in valid state.");
			_nodeRequests.remove(request);
			return null;
		} // How about waiting here for max-time, and if timeout then return null
	}

	public UnicastMessage sendNodeRequestMessage(final UnicastMessage unicastMessage) {
		return sendNodeRequestMessage(unicastMessage, 5, TimeUnit.SECONDS); // I think a default timeout maximal waiting of five seconds should be sufficient
	}

	// Small inner-class to store and manage requests and to bring together right response to right request
	class NodeRequest {
		private String _messageId;
		private CountDownLatch _latch; // Could be replaced with simple monitor-object and wait-notify, but why not use modern concurrent constructs from Java 5?
		private UnicastMessage _response;

		public NodeRequest(String messageId) {
			_messageId = messageId;
			_latch = new CountDownLatch(1);
		}

		public String getMessageId() {
			return _messageId;
		}

		public CountDownLatch getLatch() { // Better way to isolate, encapsulate this object?
			return _latch;
		}
		
		// Used by managing MachineNode to set right response to right request
		public void setResponse(UnicastMessage responseUnicastMessage) {
			_response = responseUnicastMessage;
		}

		public UnicastMessage getResponse() {
			return _response;
		}
	}
	
}
