package mygroup.myartifact.client;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import mygroup.myartifact.modules.global.supernode.webservice.Peer;

/**
 * Client talks with Web Service to register itself and to find another clients
 * Client provides a simple API to make is easy for other Software to find other
 * peers and make it possible for this peer to be found by another peers.
 * 
 * A direct connection through NAT is still a problem. But a super-node can help
 * to establish a connection. For this a "Direct-Connection Protocol" needs to
 * be introduces to have a general framework to creates a Point-to-Points
 * connection two nodes. The 'result' of a direct connection attempts is to
 * return a Socket object, representing the connection. Alternatively a
 * UDP-connection be established and a BitTorrent Protocol used to introduce
 * TCP-like features like bit rate control etc.
 * 
 * Is this still only a HTTP-Client? No(!), but a starting point to see other
 * peers or nodes, and to establish direct connections with UDP-hole punching;
 * especially when a normal attempt to establish a connection is not a trivial
 * task due to a NAT-firewall or -configuration.
 * 
 * 
 * Before this class can be used an IP needs to be set, so the client knows
 * where send the request.
 * 
 * This HttpClient communicates with a web service to get current up-to-date
 * information about current known peers.
 * 
 * PROBLEM a web service only allows pull, but no push access to resource.
 * Because of this a HttpClient needs to check periodically for new change in
 * the list of currently known peers.
 * 
 * @author daniel
 * 
 */

public class HttpClient {
	// STARTING POINT How to send various HTTP-methods
	//   http://stackoverflow.com/questions/1051004/how-to-send-put-delete-http-request-in-httpurlconnection-looks-like-not-workin
	
	// This class doesn't need a main method; only for testing the functionality
	// of the class
	//private InetAddress ip; // You know what: forget about InetAddress and switch completely to String representation for IPs?
	private String ip;
	private int port;
	
	public HttpClient() {
		// still empty, or configure with list of known peers?
		// NO, only give a single IP at runtime, and this IP also can change at
		// *any* time
	}

	//public void setIpAndPort(InetAddress ip, int port) {
	public void setIpAndPort(String ip, int port) {
		this.ip = ip;
		this.port = port;
	}
	

	// Works :)
	// Now let's go on to ... check PeersWebService to find other methods
	public List<Peer> listPeers() throws IOException {
		checkIPAndPort();
		// Make a GET-request
		String path = "/rest/peers/list"; // TODO easier way to cope with path changes here?
		InputStream is = issueGet(path);
		// Parse returned content in InputStream and pack into a convenient List of Peer objects
		// TODO switch from DOM to StAX; maybe this can reduce the lines-of-code enormously
		List<Peer> list = new ArrayList<Peer>();
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		try {
			DocumentBuilder builder = factory.newDocumentBuilder();
			Document doc = builder.parse(is);
			// try to extract XML data with internal DOM representation of XML document
			// TODO switch to StAX for less memory consumption and works much faster
			// TODO is XPath an alternative to flat DOM-traversing? Well, when I switch is StAX it doesn't matter anyway
			
			System.out.println( doc.getDocumentElement().getNodeName() ); // get name of root node, that is child of the whole document
			NodeList peers = doc.getElementsByTagName("peer"); // get alle <peer> nodes
			for(int i=0; i<peers.getLength();i++) {
				Node peerTag = peers.item(i);
				System.out.println("Durchlauf "+i);
				if(peerTag.getNodeType() == Node.ELEMENT_NODE) { // What is a ELEMENT_NODE ??
					Element e = (Element) peerTag;
					Peer peer = extractPeer(e);
					list.add(peer);
				}
			}	
		} catch (ParserConfigurationException e) {
			e.printStackTrace();
		} catch (SAXException e) {
			e.printStackTrace();
		}		
		return list;
	}
	// Helper method to extract information about a *single* peer out of in-memory DOM representation
	private Peer extractPeer(Element e) {
		// Holds complete <peer> node with all node information (id, lease, ...)
		// Why getChildNodes()? Workign in DOM-world is really inconvenient, using NodeLists even for single nodes
		// Get id
		//NodeList list = e.getElementsByTagName("id").item(0).getChildNodes(); // Walk down to TextNode of <id> node
		//Node f = (Node)list.item(0); // Grap single TextNode
		Node f = e.getElementsByTagName("id").item(0).getChildNodes().item(0);
		String id = f.getNodeValue();
		System.out.println("id ist "+id);
		
		// Get nick IF nick is present for the current node
		f = e.getElementsByTagName("nick").item(0).getChildNodes().item(0);
		String nick = null;
		if(f!=null) { // if f is null, then <nick> is *not* present; otherwise it is availbel for extraction
			nick = f.getNodeValue();
			System.out.println("nick ist "+nick);
		} else {
			System.out.println("f is null when <nick> node is *not* present");
		}
		
		// Get lease time
		f = e.getElementsByTagName("leaseTime").item(0).getChildNodes().item(0);
		long leaseTime = Long.parseLong( f.getNodeValue() ); // TODO lease is long, so convert to Long value
		
		// Get ip
		f = e.getElementsByTagName("globalIp").item(0).getChildNodes().item(0);
		String globalIp = f.getNodeValue(); // I know, handling IPs as Strings is ugly, but it's OK for the moment
		
		// TODO build a peer object of the extracted data and make it available to the external calling software
		Peer peer = new Peer(id, nick, globalIp, leaseTime);
		return peer;
	}
	// helper method to extract a *single* peer from a XML document with only information about a single peer
	private Peer extractPeer(InputStream is) throws IOException {
		Peer peer = null;
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		try {
			DocumentBuilder builder = factory.newDocumentBuilder();
			Document doc = builder.parse(is);
			peer = extractPeer( doc.getDocumentElement() ); // root tag is already peer node, so cast it and get Peer object as result
		} catch (ParserConfigurationException e) {
			e.printStackTrace();
		} catch (SAXException e) {
			e.printStackTrace();
		}
		return peer;
	}
	
	public String addPeer(Peer peer) throws IOException {
		checkIPAndPort(); // obligatory check to be sure that all needed instance variables are set
		
		// make a POST request to *create* new peer
		// PUT only needed when we want to *update* an already present peer
		
		Map<String,String> map = new HashMap<String,String>();
		// Ups: forget to pack peer information into map to built POST-data out of it
		map.put("id",peer.getId()); // id *must* always be set
		if(peer.getNick()!=null) {
			map.put("nick", peer.getNick()); // only set nick if nick ist given in peer
		}
		// TODO is it advisable to allow the client to arbitrary set the IP or even Lease-time? I don't think so!
		
		InputStream is = issuePost("/rest/peers/add", map); // here is a problem
		return extractResult(is);
	}
	
	// It's enough to supply the id of a peer; but a version which supports a fully peer is also a good idea; why not
	public String deletePeer(String id) throws IOException {
		checkIPAndPort(); // obligatory check to be sure that all needed instance variables are set
		String path = "/rest/peers/delete/"+id;
		InputStream is = issueDelete(path); // something is wrong here
		return extractResult(is); // return content of textnode of <result> tag; use helper method for this, because it is always the same step
	}
	public String deletePeer(Peer peer) throws IOException {
		return deletePeer(peer.getId());
	}
	
	// INPUT: id, OUTPUT: Peer object; I think this is a good input-output-combination
	public Peer getPeer(String id) throws IOException {
		checkIPAndPort(); // obligatory check to be sure that all needed instance variables are set
		InputStream is = issueGet("/rest/peers/get/"+id);	
		Peer peer = extractPeer(is);
		return peer;
	}
	
	// Helper method to extract text from TextNode from <result> tag
	// The XML document is created by PeerWebService
	private String extractResult(InputStream is) throws IOException {
		String result = null;
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		try {
			DocumentBuilder builder = factory.newDocumentBuilder();
			Document doc = builder.parse(is);
			result = doc.getElementsByTagName("result").item(0).getChildNodes().item(0).getNodeValue(); // Get node with result message
		} catch (ParserConfigurationException e) {
			e.printStackTrace();
		} catch (SAXException e) {
			e.printStackTrace();
		}
		return result;
		
	}
	

	// TODO split methods into general methods (GET, PUT, DELETE, POST) and
	// specific methods (listPeers, ...
	
	/* Convenient helper methods to issue GET, POST, PUT, and DELETE request on a webservice */
	private InputStream issueGet(String path) throws IOException {	
		//String requestString = "http://"+ip.getCanonicalHostName()+":"+port+path;
		String requestString = "http://"+ip+":"+port+path;
		URL url = new URL(requestString);		
		HttpURLConnection conn;
		conn = (HttpURLConnection) url.openConnection();
		conn.setRequestMethod("GET");
		// Configure request here if necessary
		conn.connect();
		InputStream is = conn.getInputStream();		
		return is; // Return as InputStream and then parse content
	}
	
	
	// TODO how to pass POST-parameters to method? With Map<String,String>? Why not, seeme like a good deal.
	private InputStream issuePost(String path, Map<String,String> params) throws IOException {
		// TODO learn how to issue a POST-request in Java 5/6
		//      have look at HttpClientTest
		
		// Build POST-data from params-map; don't forget: params could containt zero elements, or even null (don't need to check that)
		String data = "";
		// difference: a & at the beginning of Key=value
		String prefix = "";
		for(String key : params.keySet()) {
			try { data = data+prefix+URLEncoder.encode(key, "UTF-8") + "=" + URLEncoder.encode(params.get(key), "UTF-8");
			} catch (UnsupportedEncodingException e) { e.printStackTrace(); }
			prefix = "&";
		}
		
		//String requestString = "http://"+ip.getCanonicalHostName()+":"+port+path;
		String requestString = "http://"+ip+":"+port+path;
		URL url = new URL(requestString);
		HttpURLConnection conn = (HttpURLConnection) url.openConnection();
		conn.setDoOutput(true); // is this really necessary? shouldn't a POST automatically imply a doOutput=true?
		conn.setRequestMethod("POST");
		OutputStreamWriter osw = new OutputStreamWriter( conn.getOutputStream() );
		osw.write(data);
		osw.flush(); // send out data
		conn.connect();
		
		InputStream is = conn.getInputStream(); // read XML data from response; build DOM and traverse DOM-tree to get data
		// Server returned 500 as response and threw an IOException; why?
		
		return is; // return POST-result asInputStream
	}
	
	private InputStream issueDelete(String path) throws IOException {
		// I am not 100% sure about this, but the structure of a DELETE seems to be the same as for POST: you transmit POST-data
		// But isn't the correct way to transmit a HTTP-body to the server with a XML-document the right way?
		// Still have to figure this out. But for the moment: passing data like in a POST seems to be OK for the moment
		
		/* TODO TODO TODO
		 * This is kinda weird. OK, I use a POST-Param in the DELETE-method, but doesn't seem to be 
		 * the right way. I think I just need to open a URI, say I want a HTTP-DELETE, and that's it.
		 * Maybe I am going to change this. Why the heck do I need POST-mchanisms here. It just makes
		 * things a lot more complicated then they have to be!!!
		 */
		
		//String requestString = "http://"+ip.getCanonicalHostName()+":"+port+path;
		String requestString = "http://"+ip+":"+port+path;
		URL url = new URL(requestString);
		HttpURLConnection conn = (HttpURLConnection) url.openConnection();
		//conn.setDoOutput(true); // to enable sending data to the resource; Wait a minute: why to send data? My delete doesn't need any data
		//conn.setRequestProperty("Content-Type", "application/x-www-form-urlencoded"); // is this really needed for a DELETE-method? I don't think so
		conn.setRequestMethod("DELETE");
		conn.connect();
		InputStream is = conn.getInputStream(); // server returns code 500 here; why?
		return is;
	}
	
	/**
	 * Not used, ... yet.
	 */
	private InputStream issuePut() {
		// Not needed yet, but in case: http://stackoverflow.com/questions/1051004/how-to-send-put-delete-http-request-in-httpurlconnection-looks-like-not-workin
		return null;
	}
	
	// Obligatory checker method to check if IP has been already set or not
	// if not, then throw a custom IPNotSetException to remind the user to
	// *first* set a IP before using this class
	// implement as normal exception or as RuntimeException?
	// implement as RuntimeException
	class IPAndPortNotSetException extends RuntimeException {
		// How to implement a new RuntimeException?
		private static final long serialVersionUID = 1L; // I am not using any serialization mechanism, so useless in this case
	}
	
	private void checkIPAndPort() {
		if (this.ip == null || this.port == 0) // Obligatory check to see if IP has been set or
								// not
			throw new IPAndPortNotSetException();
	}

}
