package mygroup.myartifact.dispatchers;

import java.util.Properties;

import mygroup.myartifact.MachineNode;
import mygroup.myartifact.modules.messages.NodeMessage;

/* 
 * A Handler implements a functionality, e.g. chat or transfer. It is able to send and receive messages at node-level.
 * A Handler tries to proceed a message at node-level. A Handler uses the features offered by a module if needed, e.g.
 * establish a direct connection to a remote.
 * 
 * TODO The handler-mechanism could be extended to a full-featured plug-in architecture with loading handlers at runtime.
 *      It would be also possible to use SPI (META-INF/services/) or similar technology: http://stackoverflow.com/questions/1613935/java-plugin-framework-choice
 *      Or a handler doesn't introduce new functionality, but makes use of present functionality; a new handler doesn't
 *      implement a SPI, but a dispatcher makes use of an API.
 */
public abstract class Handler {
	protected MachineNode _machineNode;
	protected Properties _properties;
	protected boolean _enabled;
	
	public Handler(MachineNode machineNode, Properties properties) {
		_machineNode = machineNode;
		_properties = properties;
		_enabled = true;
	}
	
	// A handler receives message from outside and tries to handle them; returns true, if handler knows how to handle message, false otherwise
	public abstract boolean dispatchNodeMessage(NodeMessage nodeMessage); 

	// Returns a textual representation of this handler, e.g. to use in a GUI
	public abstract String getIdentifier();
	
	// Each handler may introduce new functionality offered by supplying custom methods. Method returns HandlerAction
	// objects to supply information useful to call custom methods through reflection. 
	public abstract DispatcherAction[] getActions();
		
	// A loaded handler can be enabled and disabled
	public void setEnabled(boolean enabled) {
		_enabled = enabled;
	}
	public boolean isEnabled() {
		return _enabled;
	}
}
