package mygroup.myartifact.dispatchers.chat;

import java.awt.GraphicsEnvironment;
import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Properties;

import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

import mygroup.myartifact.MachineNode;
import mygroup.myartifact.dispatchers.DispatcherAction;
import mygroup.myartifact.dispatchers.Handler;
import mygroup.myartifact.dispatchers.chat.messages.AcceptStartChat;
import mygroup.myartifact.dispatchers.chat.messages.AskToStartChat;
import mygroup.myartifact.dispatchers.chat.messages.ChatMessage;
import mygroup.myartifact.dispatchers.chat.messages.DeclineStartChat;
import mygroup.myartifact.dispatchers.chat.messages.TellMessage;
import mygroup.myartifact.dispatchers.chat.messages.TellQuitChat;
import mygroup.myartifact.modules.Node;
import mygroup.myartifact.modules.messages.NodeMessage;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A ChatHandler implements a chat functionality to exchange chat-messages between two nodes. Each chat is handled by a
 * chat-session. A ChatHandler uses a strategy to decide if a requeste for a chat-session should be accepted or
 * declined. Different predefined strategies are available, e.g. to accept/decline any chat reequest or to ask user
 * with GUI-dialog if a chat-session should be initiated.
 * Each chat-session is handled by a ChatProvider object, dealing with processing received messages and sending
 * chat-message to remote ChatHandler.
 * 
 * @author dell
 * 
 */
public class ChatHandler extends Handler {
	private Logger _logger;

	// Handles chat request received from other chat handler, decides to accept a chat session or not; uses strategy pattern
	private ChatRequestHandler _chatRequestHandler;

	// Exchanges messages with chat partner after a chat session is established; ChatProvider object represents an unique/single chat session, which is able to
	// send and receive messages. Each chatProvider communicates with a single other ChatProvider identified by ID. *Multiple* chats could take place with
	// multiple other chat dispatchers at the same time! So each accepted chat request starts a new instance of ChatProvider to chat with other node.
	private List<ChatProvider> _chatProviders;
	private Constructor<?> _chatProviderConstructor;


	public ChatHandler(MachineNode machineNode, Properties properties) {
		super(machineNode, properties);
		_logger = LoggerFactory.getLogger(ChatHandler.class);
		_chatProviders = new ArrayList<ChatProvider>();

		// Choose ChatRequestHandler, e.g. ChatRequestHandlerAlwaysDecline to decline every chat request or ChatRequestHandlerGUIDialog to choose through GUI-dialog
		if (properties.getProperty("chatRequestHandler") != null) {
			try {
				Class<? extends ChatRequestHandler> cls = (Class<? extends ChatRequestHandler>) Class.forName(properties.getProperty("chatRequestHandler"));
				Arrays.asList(cls.getInterfaces()).contains(ChatRequestHandler.class);
				Constructor<? extends ChatRequestHandler> cons = cls.getConstructor();
				this._chatRequestHandler = (ChatRequestHandler) cons.newInstance();
			} catch (ClassNotFoundException e) {
				_logger.debug("Could not load ChatRequestHandler class, using ChatRequestHandlerAlwaysAccept as default strategy implementation.");
				this._chatRequestHandler = new ChatRequestHandlerAlwaysAccept();
			} catch (SecurityException e) {
				_logger.debug("Could not instantiate ChatRequestHandler class, because SecurityManager doesn't allow this. Check and adjust settings for SecurityManager, using ChatRequestHandlerAlwaysAccept as default strategy implementation.");
				this._chatRequestHandler = new ChatRequestHandlerAlwaysAccept();
			} catch (Exception e) {
				_logger.debug("Could not instantiate ChatRequestHandler class, because no-arg constructor not available. Supply a implementation with a no-arg constructor. Using ChatRequestHandlerAlwaysAccept as default strategy implementation.");
				this._chatRequestHandler = new ChatRequestHandlerAlwaysAccept();
			} // Some kind-a ugly ExceptionHandling block, but OK for the moment, and also supplies good information to user solving any problems
		} else {
			if (GraphicsEnvironment.isHeadless()) { // No GUI available
				this._chatRequestHandler = new ChatRequestHandlerAlwaysAccept();
			} else { // GUI available, but what if testing with non user behind screen? Well, in that case, I hope you don't forget to set -chatRequestHandler parameter and choose mygroup.myartifact.dispatchers.chat.ChatRequestHandlerAlwaysAccept
				this._chatRequestHandler = new ChatRequestHandlerGUIDialog();
			}
		}

		// Choose ChatProvider, e.g. ChatProviderGUI or ChatProviderRepest, to handle chat sessions
		if (properties.getProperty("chatProvider") != null) {
			try {
				Class<? extends ChatProvider> chatProviderClass = (Class<? extends ChatProvider>) Class.forName(properties.getProperty("chatProvider"));
				_chatProviderConstructor = chatProviderClass.getDeclaredConstructor(new Class[] { ChatHandler.class, String.class }); // ChatDispatcher and StringId
			} catch (Exception e) {
				try {
					_chatProviderConstructor = ChatProviderRepeat.class.getDeclaredConstructor(new Class[] { ChatHandler.class, String.class });
				} catch (Exception e1) {
					_logger.debug("This should never happen, problems with ClassLoader? Program is in undefined state.");
				}
			}
		} else { // Use ChatProviderRepeat as fall-back of user forgot to specify a ChatProvider
			try {
				_chatProviderConstructor = ChatProviderRepeat.class.getConstructor(new Class[] { ChatHandler.class, String.class });
			} catch (Exception e) { // This will *never* happen
				_logger.debug("This should *never* happen, problem with ClassLoader? Could not create fallback class ChatProviderRepeat, program in undefined state.");
				e.printStackTrace();
			}
		}
	}
	
	// Change strategy to deal with chat requests at runtime, e.g. user switches from ChatRequestHandlerGUIDialog to ChatRequestHandlerAlwaysDecline to be not disturbed by chat requests
	public void setChatResponseListener(ChatRequestHandler requestHandler) {
		_chatRequestHandler = requestHandler;
	}

	public void removeChatProvider(ChatProvider chatprovider) {
		_chatProviders.remove(chatprovider);
	}
	
	// Used by ChatProvider to get ID of local node
	public String getLocalNodeId() {
		return _machineNode.getId();
	}
	
	// Used by ChatProvider to get identifier of remote node
	public String getPartnerIdentifier(String partnerId) {
		Node node = _machineNode.getModuleManager().getNode(partnerId);
		// There is a slightly theoretical possibility that .getNode(String-ID) could return null, usually that shouldn't be the case
		// And if so, then the application is effectively in an undefined state, so even the following !=null check won't make any big difference
		if(node!=null) {
			return node.getIdentifier();
		} else {
			return partnerId;
		}
	}
	

	// Called by ChatProvider to send chat-messages to remote ChatProvider bound to current chat-session
	public void sendChatMessage(ChatMessage chatMessage) {
		_machineNode.sendNodeMessage(chatMessage);
	}
	
	public void tryToInitiateChat(final Node remoteNode) {
		// It only makes sense establish a single chat-session between arbitrary nodes A and B, ensure this by looping through currently established chat-sessions
		for (ChatProvider provider : _chatProviders) {
			if (provider.partnerId.equals(remoteNode.getNodeId())) {
				// Chat-session between node A and B already established, useless to try to start another chat-session
				_logger.debug("Chat session with this node {} already established.", remoteNode.getIdentifier());
				if( !GraphicsEnvironment.isHeadless() ) {
					_machineNode.getModuleManager().getExecutorSerivce().submit(new Runnable() {
						@Override
						public void run() {
							JOptionPane.showMessageDialog(null, "Chat session with this node " + remoteNode.getIdentifier() + " already established.");
						}
					});
				}
				return;
			}
		}
		
		// There is no chat session established yet with remote node, so OK to try to initiate a new chat session
		// Send AskToStartChat message to this node to try to start a chat
		AskToStartChat start = new AskToStartChat(_machineNode.getId(), remoteNode.getNodeId());
		remoteNode.sendNodeMessage(start);
	}
	
	// Called by MachineNode to check if this ChatDispatcher is able to handle received NodeMessage
	@Override
	public boolean dispatchNodeMessage(final NodeMessage nodeMessage) {
		_logger.debug("Proceeding NodeMessage from remote node {}.", nodeMessage.getSender());
		// Chat dispatcher should only handle chat-messages
		if (nodeMessage instanceof ChatMessage) {
			// TODO want to switch to UDP-style messaging instead of TCP-style messaging? Could make some things easier, but no ask with Pop-up then :( let's see what future brings
			//      Currently TCP-style, user needs to accept chat-request before chat starts

			// Other chat handler wants to start chat
			if (nodeMessage instanceof AskToStartChat) {
				_logger.debug("Received AskToStartChat message from remote node {}, wants to initiate chat session.", nodeMessage.getSender());

				// Ask ChatRequestHandler if chat-request should be accepted
				Node node = _machineNode.getModuleManager().getNode(nodeMessage.getSender()); // TODO Could be null? Well, shouldn't be the case! How about introducing assert?

				// TODO is a blocking operation until user clicks on OK or CANCEL, could this be a problem? Need to analyze Thread structure, guess the WebSocket thread gets blocked by this
				// in question or when problems occur, this can/should be done more async, maybe by using future object etc.
				boolean chatAccepted = this._chatRequestHandler.wantToStartChat(node);
				if (chatAccepted) {
					_logger.debug("Chat request from remote node {} accepted, sending back AcceptStartChat message.", nodeMessage.getSender());
					AcceptStartChat accept = new AcceptStartChat(_machineNode.getId(), nodeMessage.getSender());
					_machineNode.sendNodeMessage(accept);
					// Chat accepted, start ChatProvider
					try {
						ChatProvider provider = (ChatProvider) _chatProviderConstructor.newInstance(new Object[] { this, nodeMessage.getSender() });
						_chatProviders.add(provider);
					} catch (Exception e) {
						_logger.debug("Could not create chat provider to start chat. This should never happen, program is in undefined state.");
						e.printStackTrace();
					}
				} else {
					_logger.debug("Chat request from remote node {} declined, sending back DeclineStartChat message.", nodeMessage.getSender());
					// User declined chat request: send DeclineStartChat
					DeclineStartChat decline = new DeclineStartChat(_machineNode.getId(), nodeMessage.getSender());
					_machineNode.sendNodeMessage(decline);
				}
			}

			// Other node accepted chat request
			else if (nodeMessage instanceof AcceptStartChat) {
				// Each chat session with another chat partner is handled by an instance of ChatProvider
				// How ChatProvider is chosen? How to choose between ChatProviderGUI and ChatProviderRepeat?
				_logger.debug("Remote node {} accepted previous AskToStartChat message from this node.", nodeMessage.getSender());

				try {
					String chatPartner = nodeMessage.getSender(); // Remember id of current chat partner
					ChatProvider provider = (ChatProvider) _chatProviderConstructor.newInstance(this, chatPartner);
					ChatGUI gui = new ChatGUI(provider); // Switch from ChatDispatcher to ChatProvider
					_chatProviders.add(provider);
				} catch (Exception e) {
					_logger.debug("This should never happen.");
					e.printStackTrace();
				}
			}

			// Other node declined chat request
			else if (nodeMessage instanceof DeclineStartChat) {
				_logger.debug("Remote node {} declined chat request", nodeMessage.getSender());
				if(!GraphicsEnvironment.isHeadless()) {
					SwingUtilities.invokeLater(new Runnable() {
						// Show a AlertDialog here "Node xyz declined chat request"; best to make this Box totally async in separate thread (or SwingUIThread?)
						// This way I don't need any complicated class or interface-architecture with calls; everything here is async, so showing simple Dialog-Box is OK for now
						@Override
						public void run() {
							JOptionPane.showMessageDialog(null, "Node " + nodeMessage.getSender() + " declined chat request.");
						}
					});					
				}
			}

			// Other remote node send a chat message in the context of current chat session
			else if (nodeMessage instanceof TellMessage) {
				TellMessage tellMessage = (TellMessage) nodeMessage;
				_logger.debug("Received chat message from remote node {}, chat message is {}", new Object[] { tellMessage.getSender(), tellMessage.getText() });

				// Need to route message to correct ChatProvider to handle message in current chat session context
				for (ChatProvider provider : _chatProviders) {
					if (provider.getPartnerId().equals(tellMessage.getSender())) {
						provider.messageReceived(tellMessage);
						return true;
					}
				}
				_logger.debug("Could not find appropriate ChatProvider to route TellMessage to");
			}

			// Remote node quit the chat, close chat session on local side
			else if (nodeMessage instanceof TellQuitChat) {
				_logger.debug("Remote node {} wants to quit chat.", nodeMessage.getSender());

				// Need to route message to correct ChatProvider to handle quit of chat session
				TellQuitChat tellQuitChat = (TellQuitChat) nodeMessage;
				for (ChatProvider provider : _chatProviders) {
					if (provider.getPartnerId().equals(tellQuitChat.getSender())) {
						provider.chatRemotelyQuit();
						return true;
					}
				}
				_logger.debug("Could not find appropriate ChatProvider to route TellQuitChat-message to");
			}

			return true;
		} else {
			_logger.debug("Could not handle NodeMessage from remote node {}, because type is no ChatMessage, forwarding to another dispatcher", nodeMessage.getSender());
			return false;
		}
	}

	@Override
	public String getIdentifier() {
		return "Chat";
	}

	@Override
	public DispatcherAction[] getActions() {
		DispatcherAction[] dispatcherActions = new DispatcherAction[] { new DispatcherAction("tryToInitiateChat", "Start Chat") };
		return dispatcherActions;
	}

}
