package mygroup.myartifact.dispatchers.chat;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import mygroup.myartifact.dispatchers.chat.messages.ChatMessage;
import mygroup.myartifact.dispatchers.chat.messages.TellMessage;
import mygroup.myartifact.dispatchers.chat.messages.TellQuitChat;
import mygroup.myartifact.modules.Node;

/**
 * A ChatProvider is the end-point of a chat. It receives and react on incoming chat messages, and generates messages on it's own.
 * A ChatProvider represents the chat with a single other chat dispatcher. So a ChatProdiver knows the ID of the other chat dispatcher it is chatting with.
 * 
 * ARGH this introduces another burden, ChatDispatcher needs to manage opened chats with IDs of other Chat Dispatchers; needs simple routing; well, OK for the moment
 * 
 * Chat messages are received with messageReceived(ChatMessage).
 * 
 * Chat messages are sent by generating a new ChatMessage and calling sendNodeMessage(NodeMessage) on machineNode.
 * For this a abstract class is needed, which provides protected fields for necessary object.
 * 
 */
public abstract class ChatProvider {
	private Logger logger;
	protected ChatHandler chatDispatcher;
	protected String partnerId; // Subclass and package as well, good design choice(!?!)
	protected boolean _chatClosed;
	
	public ChatProvider(ChatHandler chatDispatcher, String partnerId) {
		this.logger = LoggerFactory.getLogger(ChatProvider.class);
		this.chatDispatcher = chatDispatcher;
		this.partnerId = partnerId; // TODO is there a way to access nickname if node? make is available with ChatProvider.getIdentifier();
		this._chatClosed = false;
	}
	
	// Called by ChatDispatcher to route TellMessage to correct ChatProvider instance of current chat session context
	public String getPartnerId() {
		return this.partnerId;
	}
	
	// Called by outside ChatDispatcher to signal that ChatProvider on other remote side quit chat session
//	public abstract void chatRemotelyQuit();
	public void chatRemotelyQuit() {
		_chatClosed = true;
		// Remove from ChatDispatcher.List<ChatProvider> to really quit chat also from ChatDispatcher perspective
		chatDispatcher.removeChatProvider(this);
	}
	
	// Called by ChatProvider on this side to quit chat session; remote side gets notified about this with TellQuitChat message
	protected void quitChat() {
		// Only send TellQuitChat to remote node if chat session still running; remote node maybe already quit chat session, so sending again TellQuitChat would be obsolete and wrong
		if(!_chatClosed) {
			logger.debug("Closing chat session with node {}.", partnerId);
			TellQuitChat quitChatMessage = new TellQuitChat(chatDispatcher.getLocalNodeId(), partnerId);
			chatDispatcher.sendChatMessage(quitChatMessage);
		} else {
			logger.debug("Chat session with node {} already closed, don't need to TellQuitChat.", partnerId);
		}
		chatDispatcher.removeChatProvider(this);
	}
	
	// I think it only makes sense to route TellMessage directly to a ChatProvider; ChatDispatcher already takes care of TellQuitChat etc.
	// This class is only there for a user GUI or etc. to receive, handle messages and write appropriate responses.
	// TODO BUG FOUND why is this method never called? It should be called when TellMessage from chat partner has been send!
	public abstract void messageReceived(TellMessage tellMessage);
	
	// Here it is another picture: a chat provider should be able to quit a chat session with TellQuitChat, so here it should allowed to send general ChatMessage
	// But in the real case it will be TellMessage in over 95% of all cases of course.
	protected void sendChatMessage(ChatMessage chatMessage) {
		logger.debug("Node {} is sending ChatMessage to {}.", new Object[]{chatMessage.getSender(), chatMessage.getReceiver()});
		chatDispatcher.sendChatMessage(chatMessage);
	}
	
	// Used to send a simple chat message; wrapping into ChatMessage object is taking place in method itself
	protected void tellMessage(String msg) {
		logger.debug("Node {} is sending TellMessage to {}.", new Object[]{chatDispatcher.getLocalNodeId(), partnerId});
		TellMessage message = new TellMessage(chatDispatcher.getLocalNodeId(), partnerId, msg);
		chatDispatcher.sendChatMessage(message);
	}
	
	// Used by ChatGUI to print nice-looking nickname instead of id, which is very long due to UUID
	public String getIdentifier() {
		return chatDispatcher.getPartnerIdentifier(partnerId);
		
//		// Get nick from node with ID partnerId
//		Node node = chatDispatcher.getMachineNode().getModuleManager().getNode(partnerId);
//		
//		// There is a slightly theoretical possibility that .getNode(String-ID) could return null, usually that shouldn't be the case
//		// And if so, then the application is effectively in an undefine state, so even the following !=null check won't make any big difference
//		if(node!=null) {
//			return node.getIdentifier();
//		} else {
//			return partnerId;
//		}
	}
	
}
