package mygroup.myartifact.dispatchers.transfer;

import java.io.EOFException;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.net.Socket;
import java.util.Properties;

import mygroup.myartifact.MachineNode;
import mygroup.myartifact.dispatchers.ConnectionHandler;
import mygroup.myartifact.dispatchers.Handler;
import mygroup.myartifact.dispatchers.DispatcherAction;
import mygroup.myartifact.modules.Node;
import mygroup.myartifact.modules.messages.NodeMessage;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A TransferDispatcher is able to transfer files between local and remote node.
 * 
 * @author dell
 * 
 */
public class TransferHandler extends Handler implements ConnectionHandler {
	private Logger _logger;

	public TransferHandler(MachineNode machineNode, Properties properties) {
		super(machineNode, properties);
		_logger = LoggerFactory.getLogger(TransferHandler.class);
	}
	
	// TODO not very generic, to much tied to GUI; what is really needed is a method uploadFile or uploadFiles(File[]):boolean and return if upload went OK
	public void openUploadGUI(Node remoteNode) {
		// GUI with button to select one or more files, button to start upload of files to remote node
		TransferGUI transferGUI = new TransferGUI(this, remoteNode);
		transferGUI.setVisible(true);
	}

	@Override
	public boolean dispatchNodeMessage(NodeMessage nodeMessage) {
		return false; // Currently there are no messages a TransferDispatcher is interested in, the only thing a TransferDispatcher does is transfer files, that's all for the moment
	}

	@Override
	public String getIdentifier() {
		return "Transfer";
	}

	@Override
	public DispatcherAction[] getActions() {
		DispatcherAction[] dispatcherActions = new DispatcherAction[] { new DispatcherAction("openUploadGUI", "Upload file") };
		return dispatcherActions;
	}
	
	// This handler is able to handle direct connections (implements ConnectionHandler), needed to receive files from remote TransferHandler
	@Override
	public void handleConnection(Socket socket) {
		try {
			ObjectInputStream ois;
			try {
				InputStream is = socket.getInputStream();
				ois = new ObjectInputStream( is ); // Throws IOException
				_logger.debug("Setup ObjectInputStream");
			} catch (IOException e) {
				_logger.debug("Could not get InputStream for socket. Aborting communication.");
				e.printStackTrace();
				throw e;
			}
			
			FileInformation fileInformation = null;
			FileOutputStream fos = null;
			while (true) {
				// Each file is received in two phases: (1) receive FileInformation object to prepare file receive, (2) receive actual bytes
				try {
					_logger.debug("Waiting for FileInformation ...");
					fileInformation = (FileInformation) ois.readObject(); // Throws ClassNotFoundException, EOFException to indicate remote closed connection; I am *very* angry here, why am I *not* informed about EOFException here? Maybe because subclass of IOException
					_logger.debug("Receiving file {} with filesize of {} Byte.", new Object[]{ fileInformation.getFileName(), fileInformation.getFileSize() });
				} catch (ClassNotFoundException e) {
					_logger.debug("Remote node sends FileInformation objects to send metedata (filename, filesize) before sending actual file content. Class not found in classpath (or wrong version?). Aborting any further communication.");
					e.printStackTrace();
					throw new IOException("Could not found FileInformation class in classpath.");
				}
				// Don't know why, but catching the following creates a java.lang.IndexOutOfBoundsException
				catch (EOFException e) {
					_logger.debug("Received EOF while waiting for next FileInformation-object. Seems like remote site finished transfering files.");
					throw e;
				}
				catch (IOException e) {
					_logger.debug("Could not read FileInformation object due to arbitrary IO error. Aborting communcation.");
					e.printStackTrace();
					throw e;
				}

				File file = new File(fileInformation.getFileName());
				try {
					fos = new FileOutputStream(file); // Throws FileNotFoundException
					_logger.debug("File {} aleady exists, overwriting present content with new data.", file);
				} catch (FileNotFoundException e) {
					_logger.debug("Receiving new file {}, trying to create it.", file);
					try {
						file.createNewFile();
					} catch (IOException e1) {
						_logger.debug("Could not create new file {}. Aborting communication.");
						e1.printStackTrace();
						throw e1;
					}
				}
				
				long fileSize = fileInformation.getFileSize();
				
				// Receive file itself and safe to local file; TODO need to stop reading bytes when fileSize reached and go for next loop
				byte[] buffer = new byte[1024 * 32]; // 32 KB buffer to transfer data between InputStream and OutputStream; alternative: cash *all* data from received file to RAM, but this could led to problems with minimal RAM (embedded systems etc.)
				int bytesRead;
				int bytesReceived = 0;
				while (bytesReceived < fileSize) {
					int bytesToRead = 0;
					int length = buffer.length;
					if (fileSize - bytesReceived < length) {
						bytesToRead = (int) (fileSize - bytesReceived);
					} else {
						bytesToRead = buffer.length;
					}
					try {
						bytesRead = socket.getInputStream().read(buffer, 0, bytesToRead); // Throws IOException
						fos.write(buffer, 0, bytesRead); // Throws IOException
					} catch (IOException e) {
						_logger.debug("Could not read data from InputStream or write the data file. Aborting communication.");
						e.printStackTrace();
						throw e;
					}
					bytesReceived += bytesRead;
					System.out.println("Wanted to read " + bytesToRead + " Bytes, actually read " + bytesRead + " Bytes; aggregated Bytes received so far: " + bytesReceived);
				}
				fos.close(); // File completely transfered and wrote to hdd
			}
		} catch (EOFException e) {
			_logger.debug("Remote server closed connection, so no more files to receive. Connection in half-closed state, closing connection on this side as well to completely close connection.");
		} catch (IOException e) {
			_logger.debug("An arbitrary IO error occured, data could be corrupted, so aborting any further communication and writing to files to leave defined application state.");
			e.printStackTrace();
		} finally {
			try {
				socket.close();
			} catch (IOException e1) {
				e1.printStackTrace();
			}
		}
	}
}
