package mygroup.myartifact.modules;

import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import mygroup.myartifact.MachineNode;
import mygroup.myartifact.Utility;
import mygroup.myartifact.modules.Node.NodeNotAvailableException;
import mygroup.myartifact.modules.messages.BroadcastMessage;
import mygroup.myartifact.modules.messages.NodeMessage;
import mygroup.myartifact.modules.messages.TellNodeInfo;
import mygroup.myartifact.modules.messages.UnicastMessage;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class ModuleManager {
	private Logger _logger;
	
	private Properties _properties;
	private MachineNode _machineNode;
	private List<Module> _modules;

	private List<ModuleManagerListener> _listeners;
	private ExecutorService _executorService;

	private List<Node> _nodes; // May switch to Map<String-ID,Node> to increase performance

	public ModuleManager(Properties properties, MachineNode machineNode) {
		_logger = LoggerFactory.getLogger(ModuleManager.class);

		_listeners = new ArrayList<ModuleManagerListener>();
		_properties = properties;
		_machineNode = machineNode;
		_modules = new ArrayList<Module>();
		_nodes = new ArrayList<Node>();
		_executorService = Executors.newCachedThreadPool();
		
		if (_properties.getProperty("module") != null) {
			for (String modLine : _properties.getProperty("module").split(",")) {
				
				try {
					_logger.trace("Split is {}", modLine.split("#"));
					String mod = modLine.split("#")[0];
					int priority = Integer.parseInt(modLine.split("#")[1]);
					_logger.debug("Loading module {} with priority {}.", new Object[] { mod, priority });

					Class<?> cls = (Class<?>) Class.forName(mod);
					Class<?> superCls = cls.getSuperclass(); // TODO Loop through super-classes
					if (superCls.getName() == mygroup.myartifact.modules.Module.class.getName()) {
						Class[] types = new Class[] { Properties.class, ModuleManager.class, int.class };
						Constructor cons = cls.getConstructor(types);
						Object[] args = new Object[] { properties, this, priority };
						Module module = (Module) cons.newInstance(args);
						_modules.add(module); // Start modules with ModuleManager.start()
					}
				} catch (NumberFormatException e) {
					_logger.debug("No priority given in {}", modLine);
					_logger.debug("Correct syntax is <mod_pkg>#<priority>");
					e.printStackTrace();
				} catch (ClassNotFoundException e) {
					_logger.debug("Could not load class {}", modLine);
					e.printStackTrace();
				} catch (SecurityException e) {
					_logger.debug("Security exception occured when trying to load class.");
					e.printStackTrace();
				} catch (NoSuchMethodException e) {
					_logger.debug("Could not retrive constructor.");
					e.printStackTrace();
				} catch (IllegalArgumentException e) {
					_logger.debug("Could not call constructor.");
					e.printStackTrace();
				} catch (Exception e) {
					_logger.debug("Could not issue reflection call.");
					e.printStackTrace();
				}				
			}
		}
	}

	public void start() {
		_logger.debug("Starting modules ...");
		long startTime = System.currentTimeMillis();
		final CountDownLatch countDownLatch = new CountDownLatch(_modules.size());
		for (final Module module : _modules) {
			_executorService.submit(new Runnable() {
						@Override
						public void run() {
							module.start();
							countDownLatch.countDown();
						}
					});
		}
		try {
			countDownLatch.await(); // Wait until all modules boot up properly
			_logger.debug("All modules finished starting in {} miliseconds.", System.currentTimeMillis() - startTime);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

	public void stop() {
		// TODO may introduce same CountDownLatch waiting as in ModuleManager.start()?
		for (Module module : _modules) {
			module.stop();
		}
	}

	/*** Helper classes used by modules to access various functionality and information AND to push various data into ModuleManager like new found nodes etc. ***/
	public MachineNode getMachineNode() { // Used by modules to get information like ID, capabilities and nickname (if any, what means the String is not null)
		return this._machineNode;
	}
	
	public List<Module> getModules() { // This feels ugly: the references between MachineNode, ModuleManager, and Node; ModuleManager feels very wrong here
		return this._modules;
	}

	public ExecutorService getExecutorSerivce() { // (Mainly) used by modules to start background threads, e.g. listening for new packages etc.
		return this._executorService;
	}

	// A module uses this method to signal that is has no contact to the node with the specified ID anymore and wants to remove any references to that node
	public void removeModule(String id, Module module) {
		_logger.debug("Module {} tries to remove node with ID {}", new Object[] { module, id });
		for (Node node : _nodes) {
			if (node.getNodeId().equals(id)) {
				node.removeModule(module);
				_logger.debug("Removed Module reference {} from node {}.", new Object[] { module, id });
				return;
			}
		}
		// Couldn't find appropriate node, is this possible!?!
		_logger.debug("Could find appropriate node with ID {} to remove Module reference {} from it.", new Object[] { id, module });
	}

	// Don't give modules direct access to known nodes, don't let modules try to be smart, but only give them a way to get node to handle with it
	public Node getNode(String id, Module module) {
		// Return node reference to module; create node with id if new and didn't exist before, else return back existing node reference
		for (Node node : _nodes) {
			if (node.getNodeId().equals(id)) { // Node already exists, probably because another module already created it
				node.addModule(module);
				_logger.debug("Module {} found already known node with ID {}.", module.getIdentifier(), id);
				return node;
			}
		}

		// Could not found node with supplied ID, so create new object representing node and return reference; also notify ModuleManagerListener about new node
		_logger.debug("Don't know node with ID {}. Creating new node instance for it and adding Module {} to it.", new Object[] { id, module });
		Node newNode = new Node(id);
		newNode.addModule(module); // introduce node->AbstractModule dependency into node
		_nodes.add(newNode); // ModuleManager knows about every created/available node

		// New remote nodes doesn't know anything about this local node, Send/PUSH him all node-related information, like nickname and dispatchers with NodeMessage
		TellNodeInfo tellNodeInfo = new TellNodeInfo(_machineNode.getId(), id, new NodeInfo(_machineNode.getId(), _machineNode.getNick(), _machineNode.getEnabledDispatchersClassNames()));
		_logger.debug("Introduce to node {} by sending TellNodeInfo with message-id {}, content ID={}, Nick={}, Dispatchers={}", new Object[] { id, tellNodeInfo.getMessageId(), _machineNode.getId(), _machineNode.getNick(), _machineNode.getEnabledDispatchersClassNames() });
		sendNodeMessage(tellNodeInfo);

		_logger.debug("Informing local listeners about new found node with ID {}", id);
		for (ModuleManagerListener listener : _listeners) {
			listener.newNode(this, newNode);
		}
		return newNode;
	}

	// Use by MachineNode to update node-information
	public Node getNode(String id) {
		Node returnNode = null;
		for (Node node : _nodes) {
			if (node.getNodeId().equals(id)) {
				returnNode = node;
				break;
			}
		}
		assert returnNode != null;
		return returnNode; // Could not found node, can this ever happen? Well, shouldn't be case, but how knows, need to make some runtime tests and checks to be sure about this
	}

	public void addListener(ModuleManagerListener listener) { // Who uses this method to be notified about new nodes? JTable or NodeTableModel in MachineNodeGUI?
		if (!_listeners.contains(listener)) { // Take care of duplicates
			_listeners.add(listener);
		}
	}
	
	public void dispatchMessage(String msg) { // dispatched my submodule, give the message to MachineNode, where it can be handled appropriately, where all capabilities and the message-chain are loaded
		this._machineNode.dispatchMessage(msg);
	}
	
	public void sendNodeMessage(NodeMessage nodeMessage) throws NodeNotKnownException, NodeNotAvailableException {
		if (nodeMessage instanceof UnicastMessage) { // Send message only to a single node
			UnicastMessage unicastMessage = (UnicastMessage) nodeMessage;
			for (Node node : _nodes) {
				if (node.getNodeId().equals(unicastMessage.getReceiver())) { // Node handles everything, choosing best module and also checks if node is available (that is leaseTime!=0)
					_logger.debug("Sending unicast-NodeMessage {} to node {} with message-id {}.", new Object[] { nodeMessage.getClass().getCanonicalName(), node.getNodeId(), nodeMessage.getMessageId() });
					node.sendNodeMessage(nodeMessage);
					return; // WHY not change to true/false to signal that message was send successfully (not same as message received)... WAIT already Exceptions to signal this
				}
			}
			throw new NodeNotKnownException("Don't know node with id " + unicastMessage.getReceiver());
		} else if (nodeMessage instanceof BroadcastMessage) { // Send message to all known nodes
			_logger.debug("Is a BroadcastMessage.");
			String sender = nodeMessage.getSender();
			int counter = 0;
			List<Node> receivers = new ArrayList<Node>();
			for (Node node : _nodes) {
				if (!node.getNodeId().equals(sender)) { // Don't send message to sender
					_logger.debug("Delegating sending of NodeMessage {} to node {}.", new Object[] { nodeMessage.getClass().getCanonicalName(), node.getIdentifier() });
					node.sendNodeMessage(nodeMessage); // No need to cast to BroadcastMessage, because already did it's job as marker interface
					counter = counter + 1;
				}
			}
			_logger.debug("Sent BroadcastMessage to {} receivers: ", new Object[] { receivers.size(), receivers });
		} else { // This shoulnd't happen
			_logger.debug("Could not proceed NodeMessage, because neither UnicastMessage or BroadcastMessage. Sender is {} and NodeMessage is {}.", new Object[] { nodeMessage.getSender(), nodeMessage });
		}
	}

	/*** Used by ModuleManagerListener and ModuleManagerCaller ***/
	public ModuleManager() {
	}; // For testing purposes

	public static void main(String[] args) {
		ModuleManager manager = new ModuleManager();
	}

	public class NodeNotKnownException extends RuntimeException { // Exception
		public NodeNotKnownException(String msg) {
			super(msg);
		}
	}
}
