package mygroup.myartifact.modules;

import java.net.Socket;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import mygroup.myartifact.modules.messages.NodeMessage;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * This class holds the information of a known node. It supplies a lot public methods to easily get node information.
 * 
 * @author dell
 */
public class Node {
	private Logger _logger;

	protected final String _nodeId;
	protected String _nick;
	protected final List<Module> _modules; // A Node knows which modules found it
	
	protected List<String> _dispatchers;
	private Set<NodeListener> _listeners;

	public Node(String id) {
		this(id, null);
	}

	public Node(String id, String nick) {
		_logger = LoggerFactory.getLogger(Node.class);
		_nodeId = id;
		_nick = nick;
		_modules = new ArrayList<Module>();
		_dispatchers = new ArrayList<String>();
		_listeners = new HashSet<NodeListener>();
	}

	public String getNodeId() {
		return this._nodeId;
	}

	public String getNick() {
		return this._nick;
	}
	
	public String getIdentifier() {
		return (_nick==null) ? _nodeId : _nick;
	}

	public void setNick(String nick) {
		this._nick = nick;
		_logger.debug("Nickname of node changed to {}.", nick);
		notifyListeners();
	}
	
	public boolean isAvailable() {
		return (_modules.size()==0) ? false : true;
	}
	
	public void addModule(Module module) {
		if (!_modules.contains(module)) {
			_modules.add(module);
			_logger.debug("Node found by module {}, added to list of known modules.", module);
			notifyListeners(); // TODO only notify listeners if _modules.size() changed from 0 to 1? Went to online again, if this module was the first one (again)
		} else {
			_logger.debug("Tried to include same module {} to node again, blocking this, because makes no sense.", module);
		}
	}
	
	public void removeModule(Module module) {
		if (_modules.contains(module)) {
			_modules.remove(module);
			_logger.debug("Removed module {}.", module);
			notifyListeners();
		}
	}
	
	public void setDispatchers(List<String> dispatchers) {
		this._dispatchers = dispatchers;
		_logger.debug("List of available dispatchers for node {} changed to {}.", new Object[]{getNodeId(), getDispatchers()});
		notifyListeners();
	}

	public List<String> getDispatchers() {
		return this._dispatchers;
	}
	
	public void sendNodeMessage(NodeMessage nodeMessage) throws NodeNotAvailableException {
		if (this._modules.size() == 0) {
			_logger.debug("Cannot send NodeMessage to node {}, because no modules have this node available.", _nodeId);
			throw new NodeNotAvailableException("Node is currently not available.");
		}

		// Try to send message to this node by first choosing the best module
		Module best = null; // Here always one node available? Well, I am sure there are a lot thread-mistakes in this code to handle yet, but OK for the moment
		for (Module mod : this._modules) {
			if (best == null || mod.getPriority() < best.getPriority()) { // null-check for first loop where best is null at the beginning
				best = mod;
			}
		}
		_logger.debug("Sending {} with message-id {} to node {} through module {}.", new Object[]{nodeMessage.getClass().getCanonicalName(), nodeMessage.getMessageId(), _nodeId, best.getIdentifier()}); // getNodeId() would be equivalent to nodeMessage.getReceiver()
		best.sendNodeMessage(nodeMessage);	
	}

	/* Used by NodeListener to register itself on this node */
	public void addListener(NodeListener nodeListener) {
		this._listeners.add(nodeListener); // What about duplicate checking? Shouldn't happen, and if yes, no problem if a listener gets notified twice
	}

	private void notifyListeners() { // Notify all listeners when state of nodes changed, so call in *every* mutator method like setName(String) or decreaseLease(decrease)
		for (NodeListener listener : _listeners) {
			listener.nodeChanged(this);
		}
	}

	// Establish a direct connection handling through handler on local and remote node 
	public Socket getConnection(String dispatcherClass) {
		Collections.sort(_modules, new Comparator<Module>() { // Ordering modules by priority before-hand to ease handling
			@Override
			public int compare(Module foo1, Module foo2) {
				return foo1._priority - foo2._priority;
			}
		});
		
		Socket socket = null;
		for(Module module : _modules) {
			socket = module.getConnection(this, dispatcherClass);
			if(socket!=null) {
				return socket;
			}
		}
		return null; // No module was able to create direct connection, return null indicating that creating direct connection failed
	}
	
	// TODO introduce representation with ID, nick (if available), available handler, and modules that found this remote node
	@Override
	public String toString() {
		if(_nick==null) {
			return _nodeId;
		} else {
			return _nick;
		}
	}
	
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((_nodeId == null) ? 0 : _nodeId.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Node other = (Node) obj;
		if (_nodeId == null) {
			if (other._nodeId != null)
				return false;
		} else if (!_nodeId.equals(other._nodeId))
			return false;
		return true;
	}

	public class NodeNotAvailableException extends RuntimeException { // Exception
		public NodeNotAvailableException(String msg) {
			super(msg);
		}
	}
}
