package mygroup.myartifact.modules.global.supernode.webservice;

import org.eclipse.jetty.server.Server;
import org.eclipse.jetty.servlet.ServletContextHandler;
import org.eclipse.jetty.servlet.ServletHolder;

import com.sun.jersey.spi.container.servlet.ServletContainer;

/**
 * Class to start Jetty Webserver, setup Jersey and tell it where to look for Web-service classes.
 * Maven is used to get all dependencies (jetty, jersey, servlet-api-2.5). jersey depends on servlet-api.
 * 
 * Jersey is activated by loading a special HttpServlet into Jetty called "com.sun.jersey. ... .ServletContainer".
 * The class ServletContainer.class is loaded on startup (.setInitOrder(1)) and it reads the init-parameter
 * "com.sun.jersey.config.property.packages" to know where to search for Web-service classes.
 * It finds PeersWebService.class by annotations like @Path, @Context, @Produces and @GET(?).
 * The Web-service binds itself on the /path specified by @Path. Here it is /peers/.
 * 
 * The Web-service holds a Collection Map<String,Peer> to hold information about known nodes aka peers.
 * The collection is *not* create by the Web-service itself, but it lies in the servlet-context, a memory-space
 * shared between all Servlets in the Servlet container introduced by Jetty. The collection is stored with
 * the name "peerlist". How created "peerlist" and where it does came from? I'm glad you asked!
 * 
 * The servlet-api introduces a couple of events and interfaces to react on them. One event is fired when Jetty
 * and the servlet-context just started and is ready to hold data. The Interface is called javax.servlet.ServletContextListener
 * and it is implemented by mygroup.myartifact.modules.global.supernode.webservice.ContextListener class. This class creates
 * the collection and stores it into servlet-context.
 * Additionally the class starts a background thread to track and decrease the lease-times of known nodes and
 * removes nodes when they are not available anymore, this is when the lease time ran out or the node removed itself from the list.  
 *  
 * 
 * @author dell
 */
public class WebServiceManager {
	public boolean start() {
		// Start Jetty Webserver and Jersey Web-Service
		// gets started in start()
		
		Server server = new Server(8282);
		try { server.start(); } // Start server ...
		catch (Exception e) { e.printStackTrace(); System.err.println("IN START"); }
		try { server.join(); }  // ... and wait, not to exit main; well, not needed in my app (???)
		catch (InterruptedException e) { e.printStackTrace(); System.err.println("IN JOIN"); } 
		
		// Configure context-handler, which 'holds' all servlets and context-listeners
		final ServletContextHandler servletContext = new ServletContextHandler(server, "/", false, false); // true for session and/or security(?) handler
		boolean schedule = false;
		servletContext.addEventListener(new ContextListener(schedule)); // Startup code for 
		
		//handler.addServlet(Hello.class, "/hello"); // No reference needed to holder, yet // Hello-Servlet
		
		// WebServices: just give package name and Jersey will found *every* web service in their.
		ServletHolder restHolder = servletContext.addServlet(ServletContainer.class, "/rest/*"); // Don't understand this line, but needed for restHolder reference
		//restHolder.setInitParameter("com.sun.jersey.config.property.packages", "mygroup.myartifact"); 
		restHolder.setInitParameter("com.sun.jersey.config.property.packages", "mygroup.myartifact.modules.global.supernode.werbservice"); // Where to look for Web Services and start them; looks also subpackages? I guess so, yes.
		restHolder.setInitOrder(1); // equivalent to <load-on-startup>1</load-on-startup>
		
		return true; // Introduce test if everything's setup appropriately
	}
	
	public static void main(String[] args) { // Test method, so check if Web service can be started without any problems
		new WebServiceManager().start(); // nice feature of Java6
		
	}
}
