package mygroup.myartifact.modules.multicast;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.MulticastSocket;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketTimeoutException;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.Future;

import mygroup.myartifact.Utility;
import mygroup.myartifact.modules.Module;
import mygroup.myartifact.modules.ModuleManager;
import mygroup.myartifact.modules.Node;
import mygroup.myartifact.modules.messages.BroadcastMessage;
import mygroup.myartifact.modules.messages.NodeMessage;
import mygroup.myartifact.modules.messages.UnicastMessage;
import mygroup.myartifact.modules.multicast.messages.EmitPresence;
import mygroup.myartifact.modules.multicast.messages.MulticastMessage;
import mygroup.myartifact.modules.multicast.messages.RequestConnecting;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * 
 * TODO security and crypto; local network maybe safe, but internet is full of evil/bad people, so secure the system!
 * 
 * USEFUL INFORMATION
 * A MulticastSocket is effectively just a DatagramSocket or UDP-Socket. This leads to a very interesting and useful feature: a
 * MulticastSocket not only receives and accepts datagrams send to Multicast-IP and Port, but also *any* (UDP-)datagram addressed
 * to the same port, which uses the unique local IP of the machine. The local IP (and already known port) can be extracted from
 * any received datagram packet received from a node and should be stored in a MulticastNode field to be able to *directly*
 * address this node.
 * 
 * @author dell
 */
public class MulticastModule extends Module {
	private Logger _logger;

	// And another flaw detected, where using protected is no help: accessing protected field in chain;
	// Unfortunately needed so that ProtocolHandler can access mm from AbstractModule ... WAIT
	// Why not introduce 'protected ModuleManager getModuleManager()'? Marking this with protected is the key to restrict access to the method,
	// why not try it here and see if it leads to more robust code; at least it avoids introducing reference-copy, which obfuscates code unnecessarily
	// protected ModuleManager _mm;

	// TODO need to rework constructor to be possible to mark this fields as final
	protected InetAddress _group;
	protected Integer _port; // Setting to null possible to signal that no valid port given and joining multicast not possible

	// Why not move MulticastSocket down to ProtocolHandler? MulticastModule only creates it, but ProtocolHandler uses it all the time and needs to take care of joining or leaving MulticastGroup
//	protected MulticastSocket _multicastSocket; // Let's use protected to allow access from classes in same package, much better then providing getter methods!
	private ProtocolHandler _protocolHandler; // private Future _threadHandler; // Multicast uses (UDP-)Datagrams, which is connection-less, so lease-time mechanism needed

	private final List<MulticastNode> _multicastNodes; // Change to protected if access in ProtocolHandler needed; NO I don't think direct access is a good idea
	// Background runnable to continuously updates lease time
	private Runnable _updateLeaseTimes;
	private Future _updateLeaseTimesFuture;
	private static int _updateLeaseInterval = 1;

	// So many variables for only a single task: update lease times; how about creating new class/object holding and handling all of this? Maybe too much for the moment

	public MulticastModule(Properties properties, ModuleManager mm, int priority) { // This seems wrong, but at the same time I also need this objects
		super(properties, mm, priority);
		this._logger = LoggerFactory.getLogger(MulticastModule.class);
		_multicastNodes = new ArrayList<MulticastNode>();
		
		if ( properties.getProperty("multicastPort")!=null ) {
			try {
				_port = Integer.valueOf(properties.getProperty("multicastPort"));
				_logger.debug("Successfully read multicast listening port, is {}", _port);
				if (!(0 <= _port && _port <= 65535)) {
					throw new NumberFormatException("No valid port given");
				}
			} catch (NumberFormatException e) { // Very interesting that a NumberFormatException is a RuntimeException; this makes it tricky, because it explicitly needs to be checked; on the other hand makes programming easier
				_logger.error("No valid -mcport port number given, won't be able to start MulticastModule. Given -mcport argument is {}.", properties.getProperty("multicastPort")); // Use {} syntax to speed up logger
			}
		} else { // Module can't work without given port
			_port = null;
			_logger.error("No -mcport argument given, won't be able to start MulticastModule.");
		}
		
		// "Addresses in the range between 225.x.x.x and 232.x.x.x, and between 234.x.x.x and 238.x.x.x are available to use." http://de.wikipedia.org/wiki/Multicast: 
		if ( properties.getProperty("multicastAddress")!=null ) {
			try {
				_group = InetAddress.getByName(properties.getProperty("multicastAddress"));
				_logger.debug("Successfully read multicast listening address, is " + _group);
				// TODO how to check if given IP-Address is valid Multicast-Address?
			} catch (UnknownHostException e) {
				_logger.error("No valid address given with -mcaddress, won't be able to start MulticastModule. Value is {}", properties.getProperty("multicastAddress"));
				e.printStackTrace();
			}
		} else { // Module can't work without given address
			_group = null;
			_logger.error("No -mcaddress argument given, not be able to start MulticastModule.");
		}

		_updateLeaseTimes = new Runnable() {
			@Override
			public void run() {
				while (true) {
					try {
						Thread.sleep(_updateLeaseInterval * 1000);
						for (MulticastNode multicastNode : _multicastNodes) {
							multicastNode.decreaseLease(_updateLeaseInterval);
							if (!multicastNode.isAvailable()) { // Automatically remove MulticastNode if lease-time is zero, that is MulticastNode is not 'available' anymore or 'silent' or another good verb describing this subious state
								removeMulticastNode(multicastNode);
							}
						}
					} catch (InterruptedException e) {
						_logger.debug("Background runnable to update lease times of MulticastNodes stopped.");
						return; // MulticastModule stopped, so stop runnable as well
					}
				}
			}
		};

		_protocolHandler = new ProtocolHandler(this);
	}

	protected ModuleManager getModuleManager() {
		return this._moduleManager;
	}

	@Override
	public boolean start() {
		_logger.debug("Starting multicast module ...");
		
		if (_port == null || _group == null) {
			_logger.error("Could not start MulticastModule, because no valid -mcport and -mcaddress given.");
			return false;
		}

		// Multicast must be supported by router in local network so that Multicast can function properly; WARNING even without LAN-connection multicast may be available through local machines internal loop-interface
		if (!isMulticastAvailable()) {
			_logger.error("Could not start MulticastModule, because Multicast (IGMP) not available.");
			return false;
		}

		// Start MulticastSocket and join multicast group
		if(!_protocolHandler.start()) {
			return false;
		}
		_updateLeaseTimesFuture = _moduleManager.getExecutorSerivce().submit(_updateLeaseTimes); // Start background runnable to update lease time of known MulticastNodes
			
		return true;
	}

	private boolean isMulticastAvailable() {
		// TODO there is a method NetworkInterface.supportsMulticast, let's test if both mechanisms are robust
		// based on NetworkInterface seems to be more robust, because it is bound to a concrete interface

		// <Bind to multicast ip and port, send message and see if message is received; this works sequentially, because *every* packet is received since bind, even if thread is not blocking on receive
		_logger.debug("Checking with simple send-receive test if multicast is available...");

		// Use already created multicast socket, send datagram and try to see if anything is received at all for a short timeout; of not, multicast isn't working with high probability
		// Is it possible to use a single multicast socket object multiple times??? Let's test it!!! Even port not bound
		String msg = "Some random string";
		DatagramPacket send = new DatagramPacket(msg.getBytes(), msg.getBytes().length, _group, _port);
		MulticastSocket socket = null;
		try {
			socket = new MulticastSocket(_port);
			socket.joinGroup(_group);
			socket.send(send);
		} catch (IOException e1) {
			_logger.error("Could not check if multicast is available, because couldn't create MulticastSocket and send multicast packet. Assuming that multicast is not available");
			e1.printStackTrace();
			socket.close();
			return false;
		}

		byte[] buf = new byte[1024];
		DatagramPacket recv = new DatagramPacket(buf, buf.length);
		try {
			socket.receive(recv);
			String received = new String(buf);
			_logger.debug("Received multicast packet, assuming that multicast is available and working properly. Message was: {}", received);
			socket.close();
			return true;
		} catch (IOException e) {
			_logger.error("Could not check if multicast is available, because couldn't listen to receive multicast packet. Assuming that multicast is not available");
			e.printStackTrace();
			socket.close();
			return false;
		}
	}

	@Override
	public void stop() {
		// Stop background thread in ExecutorService handling and listening for multicast packets; retrieved Future object is entry point to handle with thread in background
		_protocolHandler.stop();
		
		// Remove all temporary MulticastNodes to tell MachineNode that those nodes are not available anymore through MulticastModule; this is not a big deal, because MulticastNodes are very temporary objects, no persistence at all
		for(MulticastNode multicastNode : _multicastNodes) {
			removeMulticastNode(multicastNode);
		}
		_updateLeaseTimesFuture.cancel(true); // Not needed anymore, because MulticastNodes removed; also important to enable grace shutdown of MachineNode
	}

	@Override
	public String getIdentifier() {
		return "Multicast";
	}
	
	@Override
	public void sendNodeMessage(NodeMessage nodeMessage) { // Used by Node to send a message with/through MulticastModule to another remote Node found with multicast
		// Send UnicastMessage to single remote MulticastNode (if representing MulticastNode object found)
		if (nodeMessage instanceof UnicastMessage) {
			UnicastMessage unicastMessage = (UnicastMessage) nodeMessage;
			
			for (MulticastNode multicastNode : _multicastNodes) {
				if (multicastNode.getId().equals(unicastMessage.getReceiver())) { // Found appropriate MulticastNode holding reference on destination Node, use it to send message to remote Node
					// multicastNode.getNode().sendNodeMessage(nodeMessage); // STOP this is wrong, Node already uses *this* ethot sendNodeMessage(NodeMessage) to send NodeMessage; this effectively introduces an endless loop!!!
					// Need to use Multicast-Datagram to send message
					String xmlData = Utility.XML.serialize(unicastMessage);
					// _protocolHandler.send(xmlData); // STOP this sends the message to *all* MulticastNodes, but I want to limit this to a single node! How about introducing new method, giving MulticastNode parameter as receiver?
					_logger.debug("Sending NodeMessage {} as unicast-message with MulticastModule to node {}.", new Object[]{ nodeMessage.getClass(), multicastNode.getId() });
					_protocolHandler.sendUnicast(xmlData, multicastNode);
					return;
				}
			}
			// Could not found appropriate MulticastNode representing remote Node
			_logger.debug("Could not found MulticastNode holding reference on Node {} to send Unicast-NodeMessage with MulticastModule.", unicastMessage.getReceiver());
			
		} else if (nodeMessage instanceof BroadcastMessage) { // Send BroadcastMessage to all nodes/multicastnodes; each MulticastMessage is automatically a broadcast-message, so no special handling needed
			String xmlData = Utility.XML.serialize(nodeMessage);
			_logger.debug("Sending NodeMessage {} as broadcast-message.", nodeMessage.getClass());
			_protocolHandler.send(xmlData);
		}
		else {
			_logger.debug("Could not send NodeMessage to remote MulticastNode, because neither UnicastMessage or BroadcastMessage. This shouldn't be the case, NodeMessage-hierarchy may be broken.");
		}
	}

	// Small helper method for ProtocolHandler to distinguish foreign from own MulticastMessages, because own messages should be ignored
	// OBSOLETE NOW AND WRONG checking if MulticastMessage is own is too late, need to check this on Datagram-Level
	// ... LET'S BRING IT BACK TO LIFE but this way the *right* way: check own MulticastMessage based on sender-ID; the same strategy is used on MachineNode-level with MachineMessages on MachineNode as well
	// 
	protected boolean ownMulticastMessage(MulticastMessage multicastMessage) {
		// Obsolete code, because takes IP to identify own MulticastMessages; this is not OK, preventing two MulticastModules communicating on the same MachineNode;
		// Better use sender-ID to compare to enable communication between two MulticastModules on same MachineNode
		
		// STOP a MulticastMessage only has an InetSocketAddress, not a Node-ID!!! So how to check!?!
		// OK now it is getting really ugly: only concrete Mesage-type like EmitPresence and EmitAbsence carry a node-ID , so I need to check each message-type separately? This is *very* error-prone when introducing new message-types
		//    but I don't see any alternative for the moment; well, I could introduce node-ID as field in each MulticastMessage, which would solve the problem immediately; is the the right decision? Don't know let's see.
		if( multicastMessage.getSenderNodeId().equals(_moduleManager.getMachineNode().getId()) ) {
			return true;
		} else {
			return false;
		}
	}
	
	// Convenient method for ProtocolHandler to handle the absence of a MulticastNode
	protected void removeMulticastNode(String nodeId) {
		
		// Would be automatically transformed to this if fa.jar placed in classpath while *compilation*, fa.jar not needed at runtime anymore; but I guess I can set '-ea' flag myself as well
		if(nodeId==null) {
			_logger.debug("Given nodeId to remove MulticastNode is null, can't handle this. Caller must always assure that parameter is !=null. Program in undefined state.");
			throw new NullPointerException("Must supply node-ID."); // Of course NullPointer and other common runtime exceptions don't need to be catched, but should indicate logical errors while runtime, where no error-handling is used
			// I hope this wan't lead to including various of NullPointer throws; well, only use for cases where normally wouldn't return nullpointer at method-call and would be silently ignored, which I don't want to happen
		}
		
		for (MulticastNode multicastNode : _multicastNodes) {
			if ( multicastNode.getId()!=null && multicastNode.getId().equals(nodeId)) { // HOW COULD I forgot to check for !=null as well? How about .equals() takes care of null? don't know; but is it even possible that MulticastNode has no null as node-ID? Don't think so, better check for nodeId parameter!
				removeMulticastNode(multicastNode);
				_logger.debug("Removed MulticastNode with ID {}.", nodeId);
				return;
			}
		}
		// This should never happen, it shouldn't be ever possible that this code gets executed
		_logger.debug("This should *never* happen, program is in undefined state. Tried to remove a MulticastNode based on given node-ID {}, but a MulticastNode with given InetSocketAddress wasn't availble.", nodeId); // _logger.debug("This should *never* happen, program is in undefined state. Tried to remove a MulticastNode based on given InetSocketAddress, but a MulticastNode with given InetSocketAddress wasn't available.");
	}

	// Internally used by MulticastModule to remove a MulticastNode
	private void removeMulticastNode(MulticastNode multicastNode) {
		// multicastNode.deleteNodeReference(); // WRONG AND OBSOLETE because Garbage Collector will automatically delete MulticastNode, even if Node still exists with reference on it, because reference tree is MulticastNode->Node
		

		// OK this is a little problem, because I removed Node-reference from MulticastNode; I needed because ModuleManager.getNode tried to send message before MulticastNode created already
		// HOW ABOUT delegating this call to ModuleManager again? ModuleManager could loop through _nodes and call removeModule on appropriate Node
		// This way I also avoid that MulticastModule tries to be smart ;) Well, but in this case this operation isn't really that smart, but OK
		// multicastNode.getNode().removeModule(this); // STOP switch to use ModuleManager to delegate removable of module
		_moduleManager.removeModule(multicastNode.getId(), this);
		_multicastNodes.remove(multicastNode);
	}

	// Called by ProtocolHandler to let MulticastModule correctly proceed EmitPresence message; create new MulticastNode if new node found etc.
	// Only used by ProtocolHandler, which is in same package, so protected modifier sufficient
	// protected void proceedEmitPresence(String id, InetSocketAddress sender) {
	protected void proceedEmitPresence(EmitPresence emitPresence) {
		String id = emitPresence.getSenderNodeId();
		InetSocketAddress sender = emitPresence.getSender();
		
		_logger.debug("Receiving EmitPresence from node {} with message-id {}.", new Object[]{ id, emitPresence.getMessageId() });
		
		// STOP why using NodeInfo in MulticastModule-world? This is *completely* wrong and could lead to misuse of supplied information;
		// TODO Failing to set Nick and Capabilities as well! OR STOP setting nick and capabilities should be the job of NodeMessages, not of MulticastMessages, so this is correct; but why nick is missing?

		// Check if a suitable MulticastNode object already exists; if so, use it; if no, create a new one: check for nodeId *and* InetSocketAddress
		for (MulticastNode multicastNode : _multicastNodes) {
			if (multicastNode.getId().equals(id) && multicastNode.getAddress().equals(sender)) { // TODO test when InetSocketAddress is equal; is this check really smart? ID should be sufficient, checking for IP+Port is obsolete and unnecessary
				// Suitable MulticastNode object exists, resetLeaseTime() on this object and finish
				_logger.debug("Resetting lease-time of MulticastNode for node {}.", multicastNode.getId());
				multicastNode.resetLease();
				return;
			}
		}

		// No suitable MulticastNode found, so create a new one with fresh lease-time set to default lease-time
		_logger.debug("No suitable MulticastNode for Node-ID {}, creating new MulticastNode with sender-address {}.", new Object[] { id, sender });

		MulticastNode multicastNode = new MulticastNode(id, sender);
		_multicastNodes.add(multicastNode); // First create+add MulticastNode, then get Node to warn remote node
		
		_logger.debug("Sending EmitPresence in reaction to received EmitPresence from node {}.", id);
		_protocolHandler.sendEmitPresence();
		
		// I think now BroadcastNodeInfo on node-level will be send without any problem; but the protocol-structure still feels very fragile, sure I need to re-think a couple of things again
		Node node = _moduleManager.getNode(id, this);
	}

	@Override
	public Socket getConnection(Node remoteNode, String dispatcherClass) {
		// Try to establish direct TCP-connection to remote node; remote node calls handleSocket on remote dispatcher if establishing connection was successful
		// remote dispatcher uses socket to get input- and outout-stream, set up receiver thread and can use output-stream to write data adhoc
		// Primary purpose: enable TransferDispatcher to implement upload of file from local node to remote node
		// 
		// I think it can be safely assumed that in a context where multicast is working without any problems, also establishing a TCP-connection should be
		// able without any problems. I think it also safe to assume that if a connection A->B is possible, then also B->A is possible, so it is not important
		// which side playe the server and client role. So testing for A->B should be sufficient; if A->B fails a test for B->A should be obsolet; well, for the moment
		
		// Start TCP-Server and listen on unused port, maybe need to search for an unused port for while with loop; let's pick high port to minimize probability from hitting already used port
		int startPort = 13444; // Start with port and find unused port incrementally
		int endPort = 13450; // Don't be too harsh on testing ports
		
		ServerSocket server = null;
		_logger.debug("Trying to establish direct connection to remote node {} in delegation of Dispatcher {}. Beginning with port {}.", new Object[]{remoteNode.getIdentifier(), dispatcherClass, startPort});
		for(int port = startPort; port<endPort && server==null; port++) {
			try {
				server = new ServerSocket(port);
				server.setReuseAddress(true);
				_logger.debug("Seccessfully bound to port {}.", port);
				
				_logger.debug("Telling remote MulticastModule {} to connect to open port {}.", port);
				RequestConnecting requestConnecting = new RequestConnecting(_moduleManager.getMachineNode().getId(), remoteNode.getNodeId(), port, dispatcherClass);
				_protocolHandler.sendMulticastMessage(requestConnecting);
				
				// Wait for connection of remote MulticastModule
				try {
					// Could THIS be the problem: a Timeout closes a connection automaitcally? I doubt that, but who knows??? But, no, the error appears *much* quicker then after 2 seconds
					server.setSoTimeout(2000); // I don't want to wait forevery, waiting for 2 seconds should be more then enough time for remote MulticastModule to answer
					Socket client = server.accept(); // OK if binding to port and accepting connection is delyed, because binding already create message queue and receives connection, which are handled later by .accept(); I LIKE HOW SOCKETS WORK :)
					_logger.debug("Remote MulticastNode successfully connected from {}:{}, returning Socket object.", new Object[]{ client.getInetAddress(), client.getPort() });
					return client;
				} catch (SocketTimeoutException e1) {
					_logger.debug("Remote MulticastNode didn't connect to opened port. Assuming arbitrary IO error, closing port.");
					server.close();
					return null;
				} catch(IOException e1) {
					_logger.debug("Arbitarry IO error occured while waiting on remote MulticastModule to connect to opened port. Closing port.");
					server.close();
					return null;
				}
			} catch (IOException e) {
				_logger.debug("Could not listen on port {}, retrying with next port {}.", new Object[]{port, ++port});
			}
		}
		
		_logger.debug("Could not found unused port, module giving up to establish direct connection.");
		return null;
	}
}
