package mygroup.myartifact.modules.multicast;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.InetSocketAddress;
import java.net.MulticastSocket;
import java.net.NetworkInterface;
import java.net.Socket;
import java.net.SocketException;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.Future;

import mygroup.myartifact.MachineNode;
import mygroup.myartifact.Utility;
import mygroup.myartifact.dispatchers.ConnectionHandler;
import mygroup.myartifact.dispatchers.Handler;
import mygroup.myartifact.modules.Node;
import mygroup.myartifact.modules.NodeInfo;
import mygroup.myartifact.modules.multicast.messages.EmitAbsence;
import mygroup.myartifact.modules.multicast.messages.EmitPresence;
import mygroup.myartifact.modules.multicast.messages.MulticastMessage;
import mygroup.myartifact.modules.multicast.messages.RequestConnecting;
import mygroup.myartifact.modules.multicast.messages.RequestEmitPresence;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A protocol handler uses a strategy to manage an up-to-date list of available nodes through Multicast.
 * 
 * @author dell
 * 
 */
public class ProtocolHandler {
	private Logger _logger;
	private MulticastModule _multicastModule;
	private static int timeout = 60;

	private MulticastSocket _multicastSocket;

	/*
	 * At least two threads are needed:
	 * 1. to receive datagram packets from other nodes
	 * 2. to send keep-alive message to update lease time on other MulticastModules
	 * Two threads should be sufficient, sending datagrams can always be send with separate method.
	 * Let's define Runnable here and use ExecutorService to get Threads for the Runnables. Stop the threads either with a Future or with a call on Socket.close()
	 */
	private Runnable _keepAliveRunnable; // Do I need this reference? Yes, if I want to be able to call start() and stop() many times in a row
	private Future _keepAliveFuture;
	private Runnable _receivePacketsRunnable;
	private Future _receivePacketsFuture; // Needed to stop Runnable, which is running with support of ExecutorService

	public ProtocolHandler(MulticastModule multicastModule) {
		_logger = LoggerFactory.getLogger(ProtocolHandler.class);
		_multicastModule = multicastModule; // Needed, to access 'protected' fields and methods in MulticastModule

		_keepAliveRunnable = new Runnable() {
			@Override
			public void run() {
				// Thread.currentThread().setName("KeepAlive"); // Useful for the beginning, because _keepAliveRunnable is long running, but when stopped misleading because other Runnable could reuse old Thread and this could produce misleading output messages
				_logger.debug("Keep-alive runnable started.");
				while (true) {
					try {
						Thread.sleep(timeout * 1000);
						_logger.debug("Sending EmitPresence to refresh lease-time.");
						sendEmitPresence();
					} catch (InterruptedException e) { // Called Future.cancel() on Future, so stopping runnable
						_logger.debug("Keep-alive runnable stopped.");
						return;
					}
				}
			}
		};

		// Läuft in eigenem Thread ab, unabhängig vom AWT-EventQueue-Thread, der beim Klicken auf Start aktiv ist/wird
		_receivePacketsRunnable = new Runnable() {
			@Override
			public void run() {
				byte[] buffer = new byte[10240]; // 1000 Byte kind-a small; may change in future; 10 KB should be fine I guess
				DatagramPacket recv = new DatagramPacket(buffer, buffer.length); // Test if reassignment can be avoided; can be avoided, only the byte-array is reassigned on each new datagram packet

				// TODO need to check with .interrupter() if .cancel(trie) was called on outside Future object, which signals to stop this Runnable
				// .receive(recv) could block for a very long time; I guess it would be a good idea to check for .interrupt() just after receive line, so that any further processing can be avoided

				try {
					while (true) {
						_logger.debug("Blocked waiting for new multicast packet.");
						_multicastSocket.receive(recv); // Blocks until next datagram received; received message is injected into 'recv', which itself injects it into 'buf'
						_logger.debug("Datagram received from {}:{} with content {}", new Object[] { recv.getAddress(), recv.getPort(), new String(buffer) });
						if (Thread.currentThread().isInterrupted()) { // TODO Is this still necessary? Down-side is a checked exception for MulticastSocket.close()
							_logger.debug("New Datagram received, but received interrupt from outside, so returning from _receivePacketsRunnable.");
							return;
						}
						
						// Why the following line is throwing a org.xml.sax.SAXParseException?
						// I think I found an error: using 'buf' as parameter seems to be wrong, it's an byte[] array and ... but a byte array is what the method wants!?!
						Object object = Utility.XML.deserialize(buffer); // Returns null if deserializing went wrong; TODO Is converting bytes with .toString to String robust? I hope so!

						if (object != null) {
							if (object instanceof MulticastMessage) {
								// Don't forget to set sender of MulticastMessage to enable easy handling of sender InetAddress and Port (=InetSocketAddress)
								MulticastMessage multicastMessage = (MulticastMessage) object;
								multicastMessage.setSender(new InetSocketAddress(recv.getAddress(), recv.getPort()));
								
								if (_multicastModule.ownMulticastMessage(multicastMessage)) {
									_logger.debug("Ignoring own MulticastMessage with message-id {}.", new Object[] { multicastMessage.getMessageId() });
								}

								// Another MulticastNode promotes it's presence with an EmitPresence message
								else if (object instanceof EmitPresence) {
									EmitPresence emitPresence = (EmitPresence) object;
									_multicastModule.proceedEmitPresence(emitPresence);
								}

								// Another MulticastModule is shutting down (stopped) and signals this with an EmitAbsence message
								else if (object instanceof EmitAbsence) {
									EmitAbsence emitAbsence = (EmitAbsence) object;
									_logger.debug("EmitAbsence received from node {} with message-id {}, removing MulticastNode.", new Object[] { emitAbsence.getSenderNodeId(), emitAbsence.getMessageId() });
									_multicastModule.removeMulticastNode(emitAbsence.getSenderNodeId()); // How about this: sender and node-ID are both Strings; what if would wrap them in custom-made classes NodeID, then I would immediately see this mistake
								}

								// Another MulticastNode wants to know how else is available, triggering every available node to send a EmitPresence message
								else if (object instanceof RequestEmitPresence) {
									RequestEmitPresence requestEmitPresence = (RequestEmitPresence) object;
									// _logger.debug("RequestEmitPresence received from {}:{} with message-id {}.", new Object[]{recv.getAddress(), recv.getPort(), requestEmitPresence.getMessageId()});
									_logger.debug("RequestEmitPresence received from node {} with message-id {}. Sending EmitPresence in response to RequestEmitPresence.", new Object[] { requestEmitPresence.getSenderNodeId(), requestEmitPresence.getMessageId() });
									// Sending an EmitPresence message is such a common task, why not make a separate method encapsulating the stuff needed to send such method
									// The same can be applied for EmitAbsence
									// _logger.debug("Sending EmitPresence in response to RequestEmitPresence.");
									sendEmitPresence();
								}

								// Another MulticastModule is trying to establish a direct connection with this MulticastModule
								else if (object instanceof RequestConnecting) {
									RequestConnecting requestConnecting = (RequestConnecting) object;

									// Only follow connecting request if this node is destination node
									if (requestConnecting.getReceiverNodeId().equals(_multicastModule.getModuleManager().getMachineNode().getId())) {
										// Try to connect to remote port *and* let handle the connection with dispatcher specified in RequestConnecting message
										// _logger.debug("RequestConnecting received from {}:{} with message-id {}.", new Object[]{recv.getAddress(), recv.getPort(), requestConnecting.getMessageId()});
										_logger.debug("RequestConnecting received from node {} with message-id {} from dispatcher {}.", new Object[] { requestConnecting.getSenderNodeId(), requestConnecting.getMessageId(), requestConnecting.getDispatcherClass() });

										// Trying to connect
										try {
											_logger.debug("Trying to connect to {}:{} ...", new Object[] { requestConnecting.getSender().getAddress(), requestConnecting.getPort() });
											final Socket socket = new Socket(requestConnecting.getSender().getAddress(), requestConnecting.getPort()); // Careful with this, throws IOException as well; I wished Eclipse would warn me about this...
											// _logger.debug("Connected, trying to forward socket-handling to dispatcher {}", requestConnecting.getDispatcherClass() );
											_logger.debug("Connected");

											// Forward handling of opened direct connection to appropriate dispatcher
											// Currently handling direct connection is bound to dispatcher, but maybe in near future there is a collection of ConnectionHandlers able to handle a connection
											// and the concrete ConnectionHandler is chosen based on identifying string (qualified class name etc.)
											List<Handler> dispatchers = _multicastModule.getModuleManager().getMachineNode().getDispatchers();
											boolean foundDispatcher = false;
											for (final Handler dispatcher : dispatchers) {
												if (dispatcher.getClass().getCanonicalName().equals(requestConnecting.getDispatcherClass())) {
													if (dispatcher instanceof ConnectionHandler) {
														// Execute in separate thread to free calling thread, so that new messages can be proceed without any huge blocking time
														foundDispatcher = true;
														_multicastModule.getModuleManager().getExecutorSerivce().submit(new Runnable() {
															// TODO could this be a problem? Is it not possible that a single socket is used and passed between different threads?
															@Override
															public void run() {
																_logger.debug("Forwarding socket-handling to dispatcher {}", dispatcher.getClass().getCanonicalName());
																((ConnectionHandler) dispatcher).handleConnection(socket);
															}
														});
														break;
													}
												}
											}
											if (!foundDispatcher) {
												// Could not found appropriate dispatcher to handle direct connection; theoretically this should *never* happen, but how knows, maybe in future Dispatchers can be loaded and unloaded ad-hoc
												_logger.debug("Could not find appropriate Dispatcher/ConnectionHandler to handle direct connection. This should never happen, application is in undefined state.");
											}
										} catch (IOException e) {
											_logger.debug("Could not establish direct connection to remote MulticastModule");
										}
									}
								}
							}

							// Message is no MulticastMessage, so past message to MachineNode, which is able to handle NodeMessage etc.
							else {
								// _logger.debug("No MulticastMessage, but anything else, e.g. NodeMessage etc. Dispatching message proceeding to MachineNode");
								_multicastModule.getModuleManager().getMachineNode().dispatchMessage(new String(buffer));
							}

						} else { // Null signals that could not deserialize with XML Bean Persistence
							// _logger.debug("Could not deserialize with XML Bean Persistence, message was {}", buf.toString());
							_logger.debug("Could not deserialize with XML Bean Persistence, message was {}", new String(buffer));
						}

						Arrays.fill(buffer, (byte) 0);// Reset byte[] to zeros, so that there is no trailing content warning/error from XMLSAXParser; this needs to be done somewhere, so why not here

					}
				} catch (SocketException e) { // Call MulticastSocket.close() to exit receive and stop module, throws SocketException in this case
					_logger.warn("MulticastSocket closed.");
					e.printStackTrace();
				} catch (IOException e) {
					_logger.warn("Could not listen for new multicast packets, trying again to log.");
					e.printStackTrace();
				}
			}
		};
	}

	public boolean start() {
		try {
			_multicastSocket = new MulticastSocket(_multicastModule._port); // Port is needed at creation time of multicast socket, so that appropriate port can be bound on OS
			_multicastSocket.joinGroup(_multicastModule._group);
			_logger.info("MulticastSocket created, bound to port and joined multicast-group {}.", new Object[] { _multicastModule._port, _multicastModule._group });
		} catch (IOException e) {
			_logger.error("Could not create multicast socket on port {} due to various IOException.");
			e.printStackTrace();
			return false;
		}

		_logger.debug("Sending EmitPresence after while starting MulticastModule");
		sendEmitPresence(); // Don't worry, already called MulticastSocket.join(group), so no datagrams will be missed

		// Start receiving and keep-alive runnable
		_receivePacketsFuture = _multicastModule.getModuleManager().getExecutorSerivce().submit(_receivePacketsRunnable);
		_keepAliveFuture = _multicastModule.getModuleManager().getExecutorSerivce().submit(_keepAliveRunnable); // To send every x seconds a EmitPresence to update lease-time on remote nodes for this node

		return true;
	}

	public void stop() {
		// Stop receiving and keep-alive runnable
		_receivePacketsFuture.cancel(true); // Send notification, runnable checks this in next loop and stops
		_keepAliveFuture.cancel(true);

		// TODO how about sending EmitAbsence message to other MulticastNodes? Good idea! How about in _receivePacketsRunnable when interrupt was executed? Not possible, is blocking since next datagram received
		// so send EmitAbsence just after .cancel(true) called on _receivePacketsRunnable so that this EmitAbsence message effectively trigger that .interrupt is checked and _receivePacketsRunnable shut down as well
		_logger.debug("Sending EmitAbsence while stopping MulticastModule");
		EmitAbsence emitAbsence = new EmitAbsence(_multicastModule.getModuleManager().getMachineNode().getId()); // SWITCH with node-ID to handle case with multiple MulticastModule within their own MachineNode run on same local machine
		String xmlData = Utility.XML.serialize(emitAbsence);
		send(xmlData); // TODO how can it be that SocketException("Socket is closed") is returned when trying to send xmlData-String here?
						// Please don't tell me that this is a bug in JVM, that closing a MulticastSocket on a specific port also closes every other on the same port ip+port listening!?!
		_multicastSocket.close(); // TODO should I use MulticastSocket.leave() first or instead of MulticastSocket.close() ??? what is going wrong here?
	}

	// Used by protocol handler to send inter-module MulticastMessages; switch to protected if MulticastModule needs to be able to send messages as well.
	protected void sendMulticastMessage(MulticastMessage multicastMessage) {
		// A general MulticastMessage has no receiver field, because message is sent to *all* nodes in this multicast context and receiver is effectively determined by MulticastModule._group and ._port
		String xmlData = Utility.XML.serialize(multicastMessage);
		_logger.debug("Sending MulticastMessage with message-id {}, content is {}", new Object[] { multicastMessage.getSenderNodeId(), xmlData });
		send(xmlData);
	}

	// Send an EmitPresence message on multicast, so every MulticastNode is receiving this message
	protected boolean sendEmitPresence() {
		MachineNode machineNode = _multicastModule.getModuleManager().getMachineNode();
		EmitPresence emitPresence = new EmitPresence(machineNode.getId());
		_logger.debug("Sending EmitPresence with message-id {}.", emitPresence.getMessageId());
		String xmlData = Utility.XML.serialize(emitPresence);
		return send(xmlData);
	}

	// Method used to send an arbitrary String (mainly XML-serialized Beans) to a single remote MulticastNode
	protected void sendUnicast(String xmlData, MulticastNode multicastNode) {
		DatagramPacket packet = null;
		try {
			NetworkInterface iface = NetworkInterface.getByInetAddress(multicastNode.getAddress().getAddress());
			if (iface != null) { // IP of multicastNode is IP of local interface on this local machine; so I need to use multicast-IP unfortunately so that datagram is received by MulticastSocket on local machine
				packet = new DatagramPacket(xmlData.getBytes(), xmlData.getBytes().length, _multicastModule._group, _multicastModule._port);
				_logger.debug("Node {} runs on local machine, sending multicast-datagram to {}:{}.", new Object[] { multicastNode.getId(), _multicastModule._group, _multicastModule._port });
			} else { // IP of multicastNode is IP of remote machinie; so I can safely use unicast-IP to send datagram only to this specific remote machine
				packet = new DatagramPacket(xmlData.getBytes(), xmlData.getBytes().length, multicastNode.getAddress().getAddress(), multicastNode.getAddress().getPort());
				_logger.debug("Node {} runs on remote machine, sending datagram to {}:{}.", new Object[] { multicastNode.getAddress().getAddress(), multicastNode.getAddress().getPort(), multicastNode.getId() });
			}
		} catch (SocketException e1) {
			packet = new DatagramPacket(xmlData.getBytes(), xmlData.getBytes().length, _multicastModule._group, _multicastModule._port);
			_logger.debug("Could not check if node {} is running on local or remote machine. Assuming node running on local machine and sending datagram as multicast-datagram to {}:{} to ensure that datagram is received.", new Object[] { _multicastModule._group, _multicastModule._port, multicastNode.getId() });
		} finally {
			try {
				_multicastSocket.send(packet);
				_logger.debug("Successfully send message to node {}.", new Object[] { multicastNode.getId() });
			} catch (IOException e) {
				_logger.debug("Failed to send Unicast-MulticastMessage to node {} due to arbitrary IOException.", new Object[] { multicastNode.getId() });
			}
		}
	}

	protected boolean send(String xmlData) {
		DatagramPacket packet = new DatagramPacket(xmlData.getBytes(), xmlData.getBytes().length, _multicastModule._group, _multicastModule._port);
		try {
			_multicastSocket.send(packet);
			return true; // Maybe caller wants to know if sending message was successful
		} catch (IOException e) {
			_logger.debug("Failed to send MulticastMessage due to arbitrary IOException, message was {}", xmlData);
			e.printStackTrace();
			return false;
		}
	}
}
