package mygroup.myartifact.modules.websocket;

import java.io.IOException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketTimeoutException;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;

import mygroup.myartifact.Utility;
import mygroup.myartifact.dispatchers.ConnectionHandler;
import mygroup.myartifact.dispatchers.Handler;
import mygroup.myartifact.modules.ModuleManager;
import mygroup.myartifact.modules.messages.NodeMessage;
import mygroup.myartifact.modules.websocket.messages.IntroduceItselfToServer;
import mygroup.myartifact.modules.websocket.messages.ReturnNodeListToClient;
import mygroup.myartifact.modules.websocket.messages.SendServerKeepAlive;
import mygroup.myartifact.modules.websocket.messages.TellClientAboutNewNode;
import mygroup.myartifact.modules.websocket.messages.TellClientShutdown;
import mygroup.myartifact.modules.websocket.messages.TellConnectToOpenedPort;
import mygroup.myartifact.modules.websocket.messages.TellToOpenPort;
import mygroup.myartifact.modules.websocket.messages.WebsocketMessage;
import mygroup.myartifact.modules.websocket.messages.WebsocketMessageResponse;

import org.eclipse.jetty.websocket.WebSocket;
import org.eclipse.jetty.websocket.WebSocketClient;
import org.eclipse.jetty.websocket.WebSocketClientFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * This class runs on the client-side of a Websocket connection and handles communication with the WebsocketServer. The
 * class takes care of protocol state and knows how to with the server, e.g. how to introduce itself or exchange state
 * about other remote nodes.
 * 
 * The protocol handler uses a strategy to supply an up-to-date list of available nodes:
 * - introduce at the beginning
 * - server pushes list of available nodes and notifications about new available nodes
 * - send BYE message when node shutdown
 * 
 * @author dell
 * 
 */
public class ProtocolHandler implements WebSocket.OnTextMessage {
	private Logger _logger;

	private ModuleManager _moduleManager;
	private WebsocketModule _websocketModule;
	private Connection _connection;
	private WebSocketClient _client;
	
	// Needed for interaction with ModuleManager
	protected Set<String> _ids;

	// Variables for protocol state
	private AtomicBoolean _identified; // This client first needs to identify itself it's ID before any other interaction
	private boolean _shutDownRequested;
	private URI _serverUri;

	// List to manage and merge request to response
	private List<WebsocketRequest> _requests;

	private Future<?> _keepAliveFuture;

	public ProtocolHandler(final WebsocketModule websocketModule, URI serverUri, ModuleManager moduleManager) {
		_logger = LoggerFactory.getLogger(ProtocolHandler.class);
		
		_ids = new HashSet<String>();
		_moduleManager = moduleManager;
		_websocketModule = websocketModule;

		_shutDownRequested = false;
		_identified = new AtomicBoolean(false);
		_serverUri = serverUri;
		
		_requests = new ArrayList<WebsocketRequest>();
	}
	
	private void send(String msg) {
		_logger.trace("WebsocketModul {} sending message to websocket server. Message is {}", new Object[] { _websocketModule.getModuleManager().getMachineNode().getIdentifier(), msg });
		try {
			this._connection.sendMessage(msg);
		} catch (IOException e) {
			_logger.trace("{} could not send message to websocket server.", _websocketModule.getModuleManager().getMachineNode().getIdentifier());
			e.printStackTrace();
		}
	}

	public void sendWebsocketMessage(WebsocketMessage websocketMessage) {
		send(Utility.XML.serialize(websocketMessage));
	}

	public WebsocketMessage sendRequestWebsocketMessage(WebsocketMessage websocketMessage, long timeout, TimeUnit unit) {
		// Basically introduce the same or similar mechanisms like with MachineNode.sendRequestNodeMessage()
		WebsocketRequest request = new WebsocketRequest(websocketMessage.getMessageId());
		_requests.add(request);
		sendWebsocketMessage(websocketMessage);
		try {
			request.getLatch().await(timeout, unit);
			_requests.remove(request);
			return request.getResponse(); // Currently it only makes sense to receive UnicastWebsocketMessage as response
		} catch (InterruptedException e) {
			_logger.debug("Waiting on CountDownLatch disturbed by interrupt. This should *never* happen. Returning null and removing NodeRequest object to bring application back in valid state.");
			_requests.remove(request);
			e.printStackTrace();
			return null;
		}
	}

	public WebsocketMessage sendRequestWebocketMessage(WebsocketMessage websocketMessage) {
		return sendRequestWebsocketMessage(websocketMessage, 5, TimeUnit.SECONDS);
	}

	public void sendNodeMessage(NodeMessage nodeMessage) {
		send(Utility.XML.serialize(nodeMessage)); // Serialize to XML and send it to wire to WebsocketServer
	}

	// Call this method to connect to web-socket server; don't forget to introduce yourself
	public void start() throws Exception {
		// Reuse existing WebSocketClient object for every reconnect
		if (_client == null) { // Just started, setup all objects to prepare first connect and all reconnects after this
			WebSocketClientFactory factory = new WebSocketClientFactory();
			factory.setBufferSize(4096);
			factory.start(); // May throw exception
			_client = factory.newWebSocketClient();
			_client.setMaxIdleTime(30000); // Need to send dummy packets to prevent Websocket-connection from going into idle state
			_client.setMaxTextMessageSize(10240); // TODO introduce message fragmenting and merging if needed
			_client.setProtocol("chat");
		}

		// Connect or reconnect to server
		Future<Connection> future = _client.open(_serverUri, this);
		
		// Throws a couple of Exceptions http://docs.oracle.com/javase/6/docs/api/java/util/concurrent/Future.html#get%28long,%20java.util.concurrent.TimeUnit%29
		future.get(10, TimeUnit.SECONDS); // Wait maximal 10 seconds to connect to server, then throw exception; the returned Connection object is also available from onOpen(Connection)

		// WebsocketServer doesn't like it when a connection becomes idle for >30 seconds and closes it; so send some noise every 15 seconds to avoid closing of connection
		// Runnable is only started here, so why not omit instance variable and paste in code directly here
		this._keepAliveFuture = _moduleManager.getExecutorSerivce().submit(new Runnable() {
			@Override
			public void run() {
				try {
					while (true) {
						Thread.sleep(15000);
						if (_identified.get()) {
							send(Utility.XML.serialize(new SendServerKeepAlive()));
						}
					}
				} catch (InterruptedException e) {
					_logger.debug("WebsocketModul {} is stopping keepalive thread.", _websocketModule.getModuleManager().getMachineNode().getIdentifier());
				}
			}
		});
		_logger.debug("WebsocketModule {} started keepalive thread.", _websocketModule.getModuleManager().getMachineNode().getIdentifier());
	}

	public void stop() {
		_logger.debug("Shutting down WebsocketModule...");
		
		_shutDownRequested = true;
		_identified.set(false);
		_keepAliveFuture.cancel(true); // Send interrupt to signal runnable to stop execution
		_connection.disconnect(); // Just close connection, no need to send BYE-message, WebsocketServer already handles this within it's onClose method (remove ClientSide object etc.)
		try {
			_client.getFactory().stop();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/* Methods for WebSocket.OnTextMessage to handle web-socket specific events */
	public void onClose(int closeCode, String message) { 
		if (_shutDownRequested) {
			_identified.compareAndSet(true, false); // .compareAndSet really needed? Is .set sufficient to use?
			_logger.debug("Connection to server closed as requested.");
		} else { // Immediately reconnect, if connection dropped unexpectedly
			// _identified=false at this stage indicates that this WebsocketModule was never able to connect to the WebsocketServer, so the server is probably down or not available at all
			// Don't try to reconnect and let WebsocketModule stay deactivated; this of course will change in the near future if a WebsocketModule knows many WebsocketServers and tries to connect to each of them
			if (!_identified.get()) {
				_logger.debug("Could not connect to given WebsocketServer URL {} at all, server probably down, not trying to reconnect, deactivating WebsocketModule.", _serverUri);
			} else { // Successfully connected, but connection dropped, so try to reconnect; effectively only a *single* try to reconnect, because _identified will be set to false again to deal with permanent server shutdown
				_identified.compareAndSet(true, false);
				int timeout = 1;
				_logger.debug("Connection to websocket server dropped unexpectedly. Trying to reestablish new connection after small timeout of {} seconds.", timeout);
				try {
					Thread.sleep(timeout * 1000);
					start(); // Try to reconnect
					_logger.debug("Successfully reconnected to websocket server.");
				} catch (Exception e) {
					_logger.debug("Could not reconnect to websocket server.");
					e.printStackTrace();
				}
			}
		}
	}

	public void onOpen(Connection conn) { // Connection just established
		_logger.debug("New connection to websocket server established.");
		_connection = conn; // Save Connection object to send message to the server
		_identified.set(true);

		// Identify with ID on server with IntroduceItselfToServer-message
		String id = _websocketModule.getModuleManager().getMachineNode().getId();
		_logger.debug("Introducing to the server with IntroduceItselfToServer-message with id {} .", id);
		IntroduceItselfToServer introduceToServer = new IntroduceItselfToServer(id);
		String xmlData = Utility.XML.serialize(introduceToServer);
		send(xmlData);
	}
	
	public void onMessage(String msg) {
		Object object = Utility.XML.deserialize(msg.getBytes()); // Could return null, indicating that could not deserialize message
		if (object == null) {
			_logger.debug("Could not deserialize message, content was {}", msg);
		} else {

			// Server sent a WebSocket-specific message, so dealing with inter-module communication
			if (object instanceof WebsocketMessage) {
				
				// Handle request-response messaging
				if (object instanceof WebsocketMessageResponse) {
					WebsocketMessageResponse websocketMessageResponse = (WebsocketMessageResponse) object;
					_logger.debug("Received WebsocketMessageResponse from node {} with message-id {} trying to reference message-id {}.", new Object[] { websocketMessageResponse.getSender(), websocketMessageResponse.getMessageId(), websocketMessageResponse.getReferencedMessageId() });
					for (WebsocketRequest websocketRequest : _requests) { // Try to associate response to right previously send request message
						if (websocketRequest.getMessageId().equals(websocketMessageResponse.getReferencedMessageId())) {
							_logger.debug("Successfully identified response to appropriate NodeRequest object, message-id is {}.", websocketMessageResponse.getReferencedMessageId());
							websocketRequest.setResponse(websocketMessageResponse); // Currently only UnicastMessage are handled as NodeResponseMessage; this may change in the near future, but there no need for this currently
							websocketRequest.getLatch().countDown(); // Signal waiting request-thread that response is available
							return; // No need to remove nodeRequest object from List, because method sendNodeRequestMessage already handles this
						}
					}
					_logger.debug("Could not associate NodeRequest referenced message id {}.", websocketMessageResponse.getReferencedMessageId());
				}

				// Server informs that a new node just entered the scene
				else if (object instanceof TellClientAboutNewNode) {
					TellClientAboutNewNode tellNewNode = (TellClientAboutNewNode) object;
					_logger.debug("Received TellClientAboutNewNode-message with ID {}.", tellNewNode.getId());
					_ids.add(tellNewNode.getId()); // Remember ID to handle module shutdown
					_websocketModule.getModuleManager().getNode(tellNewNode.getId(), _websocketModule);
					// MachineNode will take care of exchanging node-level information with remote node
				}

				// Server pushes list of currently available nodes
				else if (object instanceof ReturnNodeListToClient) {
					ReturnNodeListToClient list = (ReturnNodeListToClient) object;
					_logger.debug("Received list of available nodes with ReturnNodeListToClient-message with content {}", list.getNodeIds());
					List<String> nodeIds = list.getNodeIds();

					// Store information of each found node in Node object, retrieving those from ModuleManager
					for (String nodeId : nodeIds) {
						_ids.add(nodeId);
						_websocketModule.getModuleManager().getNode(nodeId, _websocketModule);
					}
				}

				// WebsocketServer informs that another WebsocketModule is not available anymore
				else if (object instanceof TellClientShutdown) {
					TellClientShutdown tellClientShutdown = (TellClientShutdown) object;
					String id = tellClientShutdown.getNodeId();
					_moduleManager.removeModule(id, _websocketModule);
					_ids.remove(id);
				}

				// // The following message-types deal with establishing a direct or proxy connection two nodes A and B
				// Another WebsocketModule wants to establish a direct connection, opened a local port and waits for this WebsocketModule to connect to this port
				else if (object instanceof TellConnectToOpenedPort) {
					// Node A sends his InetAddress.getLocalHost().getHostAddress() IP to node B, which should be sufficient to establish a direct connection; B tries to connect to IP:port
					final TellConnectToOpenedPort requestConnecting = (TellConnectToOpenedPort) object;
					_moduleManager.getExecutorSerivce().submit(new Runnable() { // Execute in separate thread to free calling thread, so that new messages can be proceed without blocking time
						@Override
						public void run() {
							_logger.debug("Remote node {} wants to establish direct connection, trying to connect to {}:{}", new Object[] { requestConnecting.getSender(), requestConnecting.getAddressString(), requestConnecting.getPort() });
							try {
								InetAddress address = InetAddress.getByName(requestConnecting.getAddressString());
								Socket socket = new Socket(address, requestConnecting.getPort());
								_logger.debug("Connection established, handling over socket to dispatcher {}.", requestConnecting.getDispatcherClass());

								// TODO Extract functionality in new method in ModuleManager or MachineNode, because Multicast-ProtocolHandler uses similar code
								List<Handler> dispatchers = _moduleManager.getMachineNode().getDispatchers();
								boolean foundDispatcher = false;
								for (final Handler dispatcher : dispatchers) {
									if (dispatcher.getClass().getCanonicalName().equals(requestConnecting.getDispatcherClass())) {
										if (dispatcher instanceof ConnectionHandler) {
											foundDispatcher = true;
											_logger.debug("Forwarding socket-handling to dispatcher {}", dispatcher.getClass().getCanonicalName());
											((ConnectionHandler) dispatcher).handleConnection(socket);
											break;
										}
									}
								}
								if (!foundDispatcher) {
									// Could not found appropriate dispatcher to handle direct connection; theoretically this should *never* happen, but how knows, maybe in future Dispatchers can be loaded and unloaded ad-hoc
									_logger.debug("Could not find appropriate Dispatcher/ConnectionHandler to handle direct connection. This should never happen, application is in undefined state.");
								}
							} catch (IOException e) {
								e.printStackTrace();
							}
						}
					});
				}

				// Another WebsocketModule asks this WebsocketModule to open a port, send the port number back as a response, and waits for node to connect
				else if (object instanceof TellToOpenPort) {
					final TellToOpenPort tellToOpenPort = (TellToOpenPort) object;

					_moduleManager.getExecutorSerivce().submit(new Runnable() {
						@Override
						public void run() {
							// Try to open port
							int startPort = 13440; // Start with port and find unused port incrementally
							int endPort = 13449; // Don't be too harsh on testing ports
							_logger.debug("Trying to open a local port...");
							for (int port = startPort; port < endPort; port++) {
								try {
									ServerSocket server = new ServerSocket(port);
									server.setReuseAddress(true);
									_logger.debug("Successfully bound to port {}, telling asking WebsocketModule {} to connect to this port", port);
									TellConnectToOpenedPort requestConnecting = new TellConnectToOpenedPort(_moduleManager.getMachineNode().getId(), tellToOpenPort.getSender(), port, tellToOpenPort.getDispatcherClass());
									sendWebsocketMessage(requestConnecting);

									// Wait for connection of remote WebsocketModule
									try {
										server.setSoTimeout(2000); // Wait for 2 seconds should be more then enough time for remote WebsocketModule to answer
										// TODO Wait for connection from requesting node A in separate thread, so that this thread is free to receive and handle other messages
										Socket client = server.accept();
										_logger.debug("Remote WebsocketModule successfully connected from {}:{}, handling over socket-handling to dispatcher {}.", new Object[] { client.getInetAddress(), client.getPort(), tellToOpenPort.getDispatcherClass() });
										
										List<Handler> dispatchers = _moduleManager.getMachineNode().getDispatchers(); // Give socket to handler for further interaction
										boolean foundDispatcher = false;
										for (final Handler dispatcher : dispatchers) {
											if (dispatcher.getClass().getCanonicalName().equals(requestConnecting.getDispatcherClass())) {
												if (dispatcher instanceof ConnectionHandler) {
													foundDispatcher = true;
													_logger.debug("Forwarding socket-handling to dispatcher {}", dispatcher.getClass().getCanonicalName());
													((ConnectionHandler) dispatcher).handleConnection(client);
													break;
												}
											}
										}
										if (!foundDispatcher) {
											// Could not found appropriate handler to handle direct connection; theoretically this should *never* happen, but maybe Dispatchers can be loaded and unloaded ad-hoc in the future
											_logger.debug("Could not find appropriate Dispatcher/ConnectionHandler to handle direct connection. This should never happen, application is in undefined state.");
										}
									} catch (SocketTimeoutException e1) {
										_logger.debug("Remote WebsocketModule didn't connect to opened port. Assuming arbitrary IO error, closing port.");
									} catch (IOException e1) {
										_logger.debug("Arbitarry IO error occured while waiting on remote WebsocketModule to connect to opened port. Closing port.");
									} finally {
										server.close();
									}
									break; // There should be only a *single* port tested
								} catch (IOException e) {
									_logger.debug("Could not listen on port {}, retrying with next port {}.", new Object[] { port, ++port });
								}
							}
						}
					});
				}
			}

			// Server sent anything, but a WebsocketMessage; past message to MachineNode, which is able to handle NodeMessage etc.
			else {
				_websocketModule.getModuleManager().dispatchMessage(msg);
			}
		}
	}
}
