package mygroup.myartifact.modules.websocket;

import java.io.IOException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketTimeoutException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.UnknownHostException;
import java.util.Properties;

import mygroup.myartifact.modules.Module;
import mygroup.myartifact.modules.ModuleManager;
import mygroup.myartifact.modules.Node;
import mygroup.myartifact.modules.messages.NodeMessage;
import mygroup.myartifact.modules.websocket.messages.EstablishProxyConnection;
import mygroup.myartifact.modules.websocket.messages.TellConnectToOpenedPort;
import mygroup.myartifact.modules.websocket.messages.TellOpenedPort;
import mygroup.myartifact.modules.websocket.messages.TellOpenedProxyPort;
import mygroup.myartifact.modules.websocket.messages.TellToOpenPort;
import mygroup.myartifact.modules.websocket.messages.WebsocketMessage;
import mygroup.myartifact.modules.websocket.server.WebsocketServer;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A WebsocketModule enables a MachineNode to use a connection to a WebsocketServer to find another remote nodes. All
 * protocol interaction is handled by internal ProtocolHandler object.
 * 
 * WebsocketModule communicates with WebsocketMessages to exchange basic information, e.g. node ID, to enable node-
 * level to exchange additional information on top of websocket-level.
 * 
 * @author dell
 * 
 */
public class WebsocketModule extends Module {
	private Logger _logger;
	private ProtocolHandler _protocolHandler;
	private URI _serverUri;
	
	public WebsocketModule(Properties properties, ModuleManager mm, int priority) {
		super(properties, mm, priority);
		_logger = LoggerFactory.getLogger(WebsocketModule.class);

		/*
		 * Currently there are two modi for WebsocketModule
		 * 1. create local server and connect to local server; property websocketDedicated set with port and websocketServer property ignored (if given at all)
		 * 2. connect to remote server; URL given with websocketServer property
		 * 
		 * In near future a third modi can/should be introduced, where a WebsocketModule can do both,
		 * create and connect to a local server AND connect to a remote server. So a single WebsocketModule
		 * is connected to multiple Websocket servers. This is an very important feature to cope with server
		 * failures and to be able to maintain a network of known servers, where list of known servers
		 * is spread to other nodes as well.
		 * But implementing this feature would be to complicated for the moment, save this for the near future...
		 */

		// Modus 1.: create local server and connect to this local server; ignore -wsserver <wsurl>
		if ( properties.getProperty("websocketDedicated")!=null ) {
			try {
				int port = Integer.valueOf(properties.getProperty("websocketDedicated"));
				if (!(0 <= port && port <= 65535)) {
					throw new NumberFormatException("No valid port given");
				}

				// Start local server
				WebsocketServer server = new WebsocketServer(port);
				server.start();

				// Connect to local server
				try {
					_serverUri = new URI("ws://localhost:" + port);
				} catch (URISyntaxException e) { // Should never happen, because port already checked and "ws://localhost" is a fixed value, so will always be OK
					e.printStackTrace();
				}

				_logger.debug("Dedicated configuration setup, local websocket server started.");
				return; // Finish configuration, all setup finished
			} catch (NumberFormatException e) {
				_logger.debug("Could not start local dedicated websocket server, because no valid port number given. Value is {}", properties.getProperty("multicastPort"));
			}
		}

		// Modus 2.: connect to remote server
		if (properties.getProperty("websocketServer")!=null) {
			try {
				_serverUri = new URI(properties.getProperty("websocketServer"));
				return;
			} catch (URISyntaxException e) {
				_logger.debug("Could not setup WebsocketModule, because given websocket server URI isn't well-formatted. Argument -wsserver is {}.", properties.getProperty("websocketServer"));
				_serverUri = null; // No websocketServer property given, signal this with null value
				e.printStackTrace();
			}
		}

		// TODO maybe a client is able to retrieve websocket server from another module and startup later; some lazy startup strategy could help to implement this robustly
		_logger.debug("No websocketServer property given, can't startup websocketmodule");
	}

	public ModuleManager getModuleManager() {
		return _moduleManager;
	}

	@Override
	public String getIdentifier() {
		return "Websocket";
	}

	@Override
	public boolean start() {
		_logger.debug("Starting websocket module...");
		if (_serverUri != null) {
			_protocolHandler = new ProtocolHandler(this, _serverUri, _moduleManager);
			try {
				_protocolHandler.start();
				_logger.debug("WebsocketModule Started.");
				return true;
			} catch (Exception e) {
				_logger.debug("Could not connect to websocket server.");
				e.printStackTrace();
				return false;
			}
		} else {
			_logger.debug("Could not start WebsocketModule, because no -wsserver server URI argument available.");
			return false;
		}
	}

	@Override
	public void stop() {
		_protocolHandler.stop();
		_logger.debug("WebsocketModule shutdown, cleaning up old WebsocketModule referneces to update node availability...");
		for(String id : _protocolHandler._ids) {
			_moduleManager.removeModule(id, this);
		}		
		_protocolHandler._ids.clear(); // List of ID's not needed anymore, because run finished
	}

	@Override
	public void sendNodeMessage(NodeMessage nodeMessage) {
		_protocolHandler.sendNodeMessage(nodeMessage);
	}
	
	@Override
	public Socket getConnection(Node remoteNode, String dispatcherClass) {
		/* Following configuration are possible:
		 *  (1) A and B no firewall
		 *  (2) A behind firewall, B not
		 *  (3) B behind firewall, A not (symmetric case)
		 *  (4) A behind firewall, B behind firewall <- little drawback, only works with Proxy, based achieved with proxy suppoert of Java 6
		 * The only case that can't be coped with TCP *directly* is case (4). For this UDP or SCTP can be used.
		 * An alternative is to establish a connection through a relay server, where a WebsocketServer provide relay functionality.
		 * 
		 * A couple of tests and tries are necessary until a connection can be established between node A and B. Nodes
		 * A and B should execute the tests themselves to get robust results, and not delegate this to another server or mediator node.
		 * 
		 * Currently a strategy with three steps is used:
		 *  1. A opens port and tells B to connect. A waits for connection of B a couple of seconds; if fails go to step 2. 
		 *  2. A tells B to open port. B opens port, tells A port and A tries to connect. B waits for a couple of seconds, closes port if connection couldn't be established. 
		 *  3. A tells WebsocketServer about B and to open two ports (a) and (b). WebsocketServer opens the ports and tells the port numbers to A and B.
		 *     A and B connect to the ports, WebsocketServer wires both connections. Connection through proxy established
		 * After step 2. it is very likely that both sides are behind NAT-firewalls. So WebsocketServer needs to act as relay server (not SOCKS or proxy server).
		 */
		
		_logger.debug("Trying to establish direct connection to remote node {} in delegation of dispatcher {} through WebsocketModule.", new Object[]{ remoteNode.getNodeId(), dispatcherClass, getIdentifier() });		
//		Socket socket = null;
//		_logger.debug("Phase 1: open port and tell remote to connect to opened port.");
//		if( (socket=getConnectionStep1(remoteNode, dispatcherClass))!=null ) {
//			return socket;
//		} else if( (socket=getConnectionStep2(remoteNode, dispatcherClass))!=null ) {
//			return socket;
//		} else if( (socket=getConnectionStep3(remoteNode, dispatcherClass))!=null ) {
//			return socket;
//		}
//		return null;
		
		// Isolated test relay connection through WebsocketServer
		Socket socket = null;
		if( (socket=getConnectionStep3(remoteNode, dispatcherClass))!=null ) {
			return socket;
		}
		return null;
		
	}
	
	private Socket getConnectionStep1(Node remoteNode, String dispatcherClass) {
		// 1. A opens port, tells B to connect and waits for connection a couple of seconds; go to step 2. if fails		
		int startPort = 13440; // Start with port and find unused port incrementally
		int endPort   = 13449; // Don't be too harsh on testing ports
		
		_logger.debug("Trying to establish direct connection to remote node {} in delegation of Dispatcher {}. Beginning with port {}.", new Object[]{remoteNode.getIdentifier(), dispatcherClass, startPort});
		for(int port = startPort; port<endPort; port++) {
			try {
				// Open and listen on port
				ServerSocket server = new ServerSocket(port);
				server.setReuseAddress(true);
				_logger.debug("Seccessfully bound to port {}.", port);
				
				// Telling remote node port to connect to
				_logger.debug("Telling remote WebsocketModule {} to connect to open port {}.", port);
				TellConnectToOpenedPort requestConnecting = new TellConnectToOpenedPort(_moduleManager.getMachineNode().getId(), remoteNode.getNodeId(), port, dispatcherClass);
				_protocolHandler.sendWebsocketMessage(requestConnecting);
				
				// Wait for connection of remote WebsocketModule
				try {
					server.setSoTimeout(2000); // Waiting for 2 seconds should be more then enough time for remote MulticastModule to answer
					Socket client = server.accept();
					_logger.debug("Remote WebsocketModule successfully connected from {}:{}, returning Socket object.", new Object[]{ client.getInetAddress(), client.getPort() });
					return client;
				} catch (SocketTimeoutException e1) {
					_logger.debug("Remote WebsocketModule didn't connect to opened port. Assuming arbitrary IO error, closing port.");
					server.close();
					return null;
				} catch(IOException e1) {
					_logger.debug("Arbitarry IO error occured while waiting on remote WebsocketModule to connect to opened port. Closing port.");
					server.close();
					return null;
				}
			} catch (IOException e) {
				_logger.debug("Could not listen on port {}, retrying with next port {}.", new Object[]{port, ++port});
			}
		}
		return null;
	}
	
	private Socket getConnectionStep2(Node remoteNode, String dispatcherClass) {
		// 2. A tells B to open port. B opens arbitrary port, tells A port and A tries to connect. B waits for a couple of seconds, closes port if connection not established.
		TellToOpenPort tellToOpenPort = new TellToOpenPort();
		WebsocketMessage response = _protocolHandler.sendRequestWebocketMessage(tellToOpenPort); // Send as request to receive opened port as response
		if(response!=null) {
			TellOpenedPort tellOpenedPort = (TellOpenedPort) response;
			try {
				Socket socket = new Socket(tellOpenedPort.getInetAddress(), tellOpenedPort.getPort());
				return socket;
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		return null;
	}
	
	private Socket getConnectionStep3(Node remoteNode, String dispatcherClass) {
		// 3. A tells WebsocketServer about B and to open two ports X and Y. WebsocketServer opens the ports and tells the port number to A and B.
		//    A and B connect to the ports, WebsocketServer wires both connections. Connection through proxy established		
		EstablishProxyConnection establishProxyConnection = new EstablishProxyConnection(remoteNode.getNodeId(), dispatcherClass);
		WebsocketMessage response = _protocolHandler.sendRequestWebocketMessage(establishProxyConnection);
		if(response!=null) {
			TellOpenedPort tellOpenedPort = (TellOpenedPort) response;
			_logger.debug("Received TellOpenedPort-response from WebsocketServer, opened port {}, trying to connect...", tellOpenedPort.getPort());
			try {
				Socket server = new Socket(InetAddress.getByName(_serverUri.getHost()).getHostAddress(), tellOpenedPort.getPort());
				_logger.debug("Successfully connected to {}:{}.", new Object[]{ InetAddress.getByName(_serverUri.getHost()).getHostAddress(), tellOpenedPort.getPort()});
				return server;
			} catch (UnknownHostException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		return null;
	}
}
