package mygroup.myartifact.modules.websocket.server;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.CountDownLatch;

import mygroup.myartifact.Utility;
import mygroup.myartifact.modules.Node;
import mygroup.myartifact.modules.messages.BroadcastMessage;
import mygroup.myartifact.modules.messages.NodeMessage;
import mygroup.myartifact.modules.messages.UnicastMessage;
import mygroup.myartifact.modules.websocket.messages.EstablishProxyConnection;
import mygroup.myartifact.modules.websocket.messages.IntroduceItselfToServer;
import mygroup.myartifact.modules.websocket.messages.ReturnNodeListToClient;
import mygroup.myartifact.modules.websocket.messages.SendServerKeepAlive;
import mygroup.myartifact.modules.websocket.messages.TellClientAboutNewNode;
import mygroup.myartifact.modules.websocket.messages.TellConnectToOpenedPort;
import mygroup.myartifact.modules.websocket.messages.TellOpenedPort;
import mygroup.myartifact.modules.websocket.messages.TellProxyConnectionFailed;
import mygroup.myartifact.modules.websocket.messages.UnicastWebsocketMessage;
import mygroup.myartifact.modules.websocket.messages.WebsocketMessage;

import org.eclipse.jetty.websocket.WebSocket;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A ClientSide object runs on the WebsocketServer and represents the client-side of a Websocket connection. The server
 * receives arbitrary messages from the client. The server PUSHes-messages to the client to inform about available nodes.
 * 
 * @author dell
 */
public class ClientSide implements WebSocket.OnTextMessage {
	private Logger _logger;
	
	private String _id;
	private InetAddress _remoteIp;
	private WebsocketServer _websocketServer;
	
	private boolean _identified;
	private Connection _connection;
	
	public ClientSide(WebsocketServer server, InetAddress remoteIp) {
		_logger = LoggerFactory.getLogger(ClientSide.class);
		this._websocketServer = server;
		_remoteIp = remoteIp; // Use remote IP to tell WebsocketModule IP of remote node, so that this information can be encapsulated in WebsocketNode object and later used to establish direct connection
	}
	
	@Override
	public void onOpen(Connection connection) {
		_connection = connection; // Remote client connected to server; the clients first needs to identify itself
	}
	
	@Override
	public void onClose(int closeCode, String message) {
		// Client closed connection or connection dropped unexpectedly; client may reconnect if necessary
		// Notify other WebsocketModules that client is not available anymore; WebsocketServer takes care of removing ClientSide object
		_websocketServer.notifyWebsocketModuleNotAvailableAnymore(_id, this);
	}
	
	@Override
	public void onMessage(String msg) {
		Object object = Utility.XML.deserialize(msg.getBytes()); // Returns null if failed to deserializing message

		// WebSocket client first needs to identify itself with IntroduceItselfToServer message; until this point WebsocketServer will refuse to proceed or route any messages between WebsocketModules
		if (!_identified) {
			if (object instanceof IntroduceItselfToServer) { // This is a WebsocketMessage
				IntroduceItselfToServer tellId = (IntroduceItselfToServer) object;
				_id = tellId.getId();
				_identified = true;
				_logger.debug("WebsocketModule with ID {} successfully identified itself.", tellId.getId());
				
				// Let WebsocketServer PUSH data about available nodes into WebsocketModule just after WebsocketModule identified itself
				List<String> nodeIds = new ArrayList<String>();
				for (ClientSide client : _websocketServer.getClients()) {
					if( client.getNodeId()!=null && !client.getNodeId().equals(_id) ) { // Node behind this ClientSide doesn't need to know information about itself
						nodeIds.add(client.getNodeId());
					}
				}
				ReturnNodeListToClient list = new ReturnNodeListToClient(nodeIds);
				_logger.debug("Telling WebsocketModule with ID {} current list of available nodes as {}", new Object[]{tellId.getId(), nodeIds});
				send(Utility.XML.serialize(list)); // Send list of available nodes to client
								
				// Tell other clients about new node with node-ID; that way the client has the chance to tell this new node information about itself on node-level (nickname and dispatchers)
				TellClientAboutNewNode tellClient = new TellClientAboutNewNode(_id, _remoteIp); // TODO here would be a good place to introduce to WebsocketServer visible IP of remote node to other node
				String xmlData = Utility.XML.serialize(tellClient);
				for(ClientSide client : _websocketServer.getClients()) {
					if(client != this) { // send this message to all clients, but this client
						client.send(xmlData);
					}
				}
			} else {
				_logger.debug("WebsocketModule {} tried to send message before identifying itself. Blocking any interaction and refusing routing any messages until client identified itself.", _id);
			}
		}
		
		// Client sent a WebSocket-specific message to this server, so dealing with inter-module communication between WebsocketModule and WebsocketServer
		else if (object instanceof WebsocketMessage) {
			
			// Client sent a keep-alive-message
			if(object instanceof SendServerKeepAlive) {
				// No further processing needed
			}
			
			// WebsocketModule A wants to establish a relay connection through this WebsocketServer to another WebsocketModule B
			else if (object instanceof EstablishProxyConnection) {
				final EstablishProxyConnection establishProxyConnection = (EstablishProxyConnection) object;
				_logger.debug("Node {} wants to establish relay connection with node {}.", new Object[]{_id, establishProxyConnection.getDestinationNodeId()});
				// Open ports (a) and (b), tell A port (a) and tell B port(b), wait for both to connect
				_websocketServer.getExecutorService().submit(new Runnable() {
					@Override
					public void run() {
						// Try to open two ports
						int startPort = 20000, endPort = 20010;
						int porta = 0, portb = 0;
						ServerSocket server1=null, server2=null;
						_logger.debug("Trying to open two ports...");
						for(int i=startPort; i<endPort; i++) {
							try {
								server1 = new ServerSocket(i);
								porta = i;
								break;
							} catch (IOException e) {
								_logger.debug("Could not bind port {}, trying next port...", i);
							}
						}
						for(int i=porta+1; i<endPort; i++) {
							try {
								server2 = new ServerSocket(i);
								portb = i;
								break;
							} catch (IOException e) {
								_logger.debug("Could not bind port {}, trying next port...", i);
							}
						}
						
						if(porta!=0&&portb!=0) {
							_logger.debug("Opened ports {} and {}.", new Object[]{porta, portb});
							// Tell A and B opened ports
							_logger.debug("Telling node {} to connect to port {}.", new Object[]{_id, porta});
							TellOpenedPort tellOpenedPort = new TellOpenedPort(null, null, establishProxyConnection.getMessageId(), null, porta); // TODO introduce message-type TellConnectToOpenedPortResponse to not misuse other message-type
							send(Utility.XML.serialize(tellOpenedPort));
							
							TellConnectToOpenedPort tellConnectToOpenedPortB = new TellConnectToOpenedPort(null, establishProxyConnection.getDestinationNodeId(), portb, establishProxyConnection.getDispatcherClass());
							// TODO Handle case where ClientSide with given receiver-string is not available
							for (ClientSide client : _websocketServer.getClients()) {
								if( client.getNodeId()!=null && client.getNodeId().equals(establishProxyConnection.getDestinationNodeId()) ) { // Only deal with already introduced clients/nodes
									_logger.debug("Telling node {} to connect to port {}.", new Object[]{establishProxyConnection.getDestinationNodeId(), portb});
									client.send(Utility.XML.serialize(tellConnectToOpenedPortB));
									break;
								}
							}
							
							// Wait for connection from node A and B; need two separate runnable for this
							_logger.debug("Waiting for both nodes to connect ...");
							CountDownLatch latch = new CountDownLatch(2);
							HandleClientConnection handler1 = new HandleClientConnection(server1, latch);
							HandleClientConnection handler2 = new HandleClientConnection(server2, latch);
							handler1.setHandler(handler2);
							handler2.setHandler(handler1);
							_websocketServer.getExecutorService().submit(handler1);
							_websocketServer.getExecutorService().submit(handler2);
						} else {
							_logger.debug("Could not open two ports.");
							// Could not bind two ports, failed to establish relay connection; tell node A about this; messages travels directly from WebsocketServer to WebsocketModule, so sender and receiver not needed 
							TellProxyConnectionFailed failed = new TellProxyConnectionFailed(null, null, establishProxyConnection.getMessageId());
							send(Utility.XML.serialize(failed));
							
							// Close ports if opened to free resources
							try {
								server1.close();
							} catch (IOException e) {
							}
							try {
								server2.close();
							} catch (IOException e) {
							}
						}					
					}
				});
			}
			
			// Client wants to send UnicastWebsocketMessage to another WebsocketModule; just re-route the message to the right receiver
			else if(object instanceof UnicastWebsocketMessage) {
				UnicastWebsocketMessage unicastWebsocketMessage = (UnicastWebsocketMessage) object;
				String receiver = unicastWebsocketMessage.getReceiver();
				
				for (ClientSide client : _websocketServer.getClients()) {
					if( client.getNodeId()!=null && client.getNodeId().equals(receiver) ) { // Only deal with already introduced clients/nodes
						client.send(msg);
						break;
					}
				}
				// No introduced WebsocketModule with given receiver-String available; how to react on such situation?
			}
			
			else {
				_logger.debug("Can't handle message {}", object);
			}
		}
		
		// Client sent a NodeMessage; a WebSocket server isn't interpreting this messages itself, but only routing them to the appropriate destination
		else if(object instanceof NodeMessage) {
			
			// An UnicastMessage is sent to a single destination
			if(object instanceof UnicastMessage) {
				UnicastMessage unicastMessage = (UnicastMessage) object;
				_websocketServer.sendUnicastMessage(unicastMessage);	
			}
			
			// A BroadcastMessage is sent to every client (of course not to the client that sent this message)
			else if(object instanceof BroadcastMessage) {
				BroadcastMessage broadcastMessage = (BroadcastMessage) object;
				_websocketServer.sendBroadcastMessage(broadcastMessage);
			}
		}

		else {
			_logger.debug("WebsocketServer could not understand message from WebsocketModule {}, doesn't know how to handle. Message was {}", new Object[]{_id, msg});
		}
	}
	
	public String getNodeId() {
		return _id;
	}
	
	public void send(String msg) { // Method used to send serialized messages to WebsocketModule behind this ClientSide instance
		try {
			_logger.debug("WebsocketServer sends routed message to WebsocketModule {}, message is {}", new Object[] { _id, msg });
			_connection.sendMessage(msg);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
}
