package mygroup.myartifact.modules.websocket.server;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import javax.servlet.http.HttpServletRequest;

import mygroup.myartifact.Utility;
import mygroup.myartifact.modules.ModuleManager;
import mygroup.myartifact.modules.Node;
import mygroup.myartifact.modules.messages.BroadcastMessage;
import mygroup.myartifact.modules.messages.UnicastMessage;
import mygroup.myartifact.modules.websocket.messages.TellClientShutdown;

import org.eclipse.jetty.server.Handler;
import org.eclipse.jetty.server.Server;
import org.eclipse.jetty.server.handler.ResourceHandler;
import org.eclipse.jetty.websocket.WebSocket;
import org.eclipse.jetty.websocket.WebSocketHandler;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class WebsocketServer {
	private Logger _logger;
	
	private Server _server;
	private int _port;
	
	private ExecutorService _executorService;
	
	private List<ClientSide> _clients;
	private List<String> _ids;
	
	public WebsocketServer(int port) {
		_logger = LoggerFactory.getLogger(WebsocketServer.class);
		
		_ids = new ArrayList<String>();
		_clients = new ArrayList<ClientSide>();
		_port = port;
		
		_executorService = Executors.newCachedThreadPool();		
	}
	
	public ExecutorService getExecutorService() {
		return _executorService;
	}
	
	public void sendUnicastMessage(UnicastMessage unicastMessage) {
		String receiver = unicastMessage.getReceiver();
		ClientSide client = getClientSide(receiver);
		if(client!=null) {
			client.send(Utility.XML.serialize(unicastMessage));
		} else {
			_logger.debug("Node with ID {} not available Could not send");
		}
	}
	
	public void sendBroadcastMessage(BroadcastMessage broadcastMessage) {
		String sender = broadcastMessage.getSender();
		ClientSide senderClient = getClientSide(sender);
		for(ClientSide client : _clients) {
			if(senderClient!=client) { // Send message to all clients except to sender of this broadcast message
				client.send(Utility.XML.serialize(broadcastMessage));
			}
		}
	}
	
	public void notifyWebsocketModuleNotAvailableAnymore(String id, ClientSide clientSide) {
		// Notify WebsocketModules that node is not available anymore
		for(ClientSide client : _clients) {
			if(!client.getNodeId().equals(id)) {
				TellClientShutdown tellClientShutdown = new TellClientShutdown(id);
				client.send(Utility.XML.serialize(tellClientShutdown));	
			}
		}
		_clients.remove(clientSide);
	}
	
 	public List<ClientSide> getClients() {
		return this._clients;
	}
 	
 	// Convenient method to get the ClientSide for a specified nodeId; return null if no Node with given ID is bound to this ClientSide
 	private ClientSide getClientSide(String id) {
 		for(ClientSide client : _clients) {
 			String nodeId = client.getNodeId();
 			if(nodeId!=null&&nodeId.equals(id)) {
 				return client;
 			}
 		}
 		return null;
 	}
 	
	public boolean start() {		
		final WebsocketServer thisServer = this;		
		WebSocketHandler websocketHandler = new WebSocketHandler() {
			public WebSocket doWebSocketConnect(HttpServletRequest request, String protocol) {
				_logger.debug("Client tries to initiate websocket connection ...");
				try {
					InetAddress remoteIp = InetAddress.getByName(request.getRemoteAddr());
					ClientSide clientWebsocket = new ClientSide(thisServer, remoteIp);
					_clients.add(clientWebsocket);
					return clientWebsocket;
				} catch (UnknownHostException e) { // Should *never* be the case, and if so print a warning and close the connection by returning null
					_logger.debug("Could not create Websocket-connection because could not parse remote IP to InetAddress object, remote IP string was {}.", request.getRemoteAddr());
					e.printStackTrace();
					return null;
				}
			}
		};

		// Build chain of handlers: server -> websocketHandler -> resourceHandler; That way a web-socket connection initiated by a client can be seen and executed
		_logger.debug("Building Jetty handler chain");
		_server = new Server(_port);
		_server.setHandler(websocketHandler);
		Handler resourceHandler = new ResourceHandler();
		websocketHandler.setHandler(resourceHandler);
		try {
			_server.start();
			_logger.debug("Started Jetty web server on port {} and ready to handle websocket client connections.", _port);
			return true;
		} catch (Exception e) {
			_logger.debug("Could not start Jetty web server on port {}.", _port);
			e.printStackTrace();
			return false;
		}
	}

	public void stop() {
		try {
			_server.stop();
			_logger.debug("Stopped Jetty web server on port {}.", _port);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}
